/*  Project: Open Source Gmini Tools
    file: CPUCore.cpp, abstraction of the TCC730 cores

    Copyright (C) 2004  Michael R. Donat <michael@donat.org>
         with contributions by gromit' and Treyqae

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
using namespace std;
#include <assert.h>
#include <stdio.h>
#include "types.h"
#include "MMU.h"
#include "decodeOP.h"
#include "formatOP.h"
#include "BreakpointList.h"
#include "CPUCore.h"

const char CPUCore::errcodestr[14][20] = {
   "SUCCESS", "NOTIMPLEMENTED", "INVALID", "RESERVED",
   "OPSIZE", "MEMORY", "ABSJUMP", "RELJUMPUP", "RELJUMPDN",
   "READONLY", "WRITEONLY", "KNOWNDATA", "OOPS", "BREAKPOINT" };
CPUCore::CPUCore( MMU *mem ) {
	memory = mem;
	depth = olddepth = 0;
	for( int i = 0; i < 283; ++i )
		usage[i] = 0;
	thisop = lastop = 0;
	steps = 0;
	oldR[0] = R[0] = R[1] = R[ 2] = R[ 3] = R[ 4] = R[ 5] = R[ 6] = R[ 7] = 
	          R[8] = R[9] = R[10] = R[11] = R[12] = R[13] = R[14] = R[15] = 
	          E[0] = E[1] = E[2] = E[3] = E[4] = E[5] = E[6] = E[7] = 0;
	PCC = SR = 0;
	SPC_FIQ = SPC_IRQ = 0;
	SPSR[0] = SPSR[1] = SPSR[2] = SPSR[3] = SPSR[4] = SPSR[5] = SPSR[6] = 0;
	verify();
}
void CPUCore::locatefunctions() {
	u8 signature[13] = { 0x10, 0x7F, 0xC8, 0x09, 0xAD, 0x1B, 0xAB, 0x41, 0xAC, 0x1A, 0xAA, 0x41, 0x9F };
	//u8 signature[7] = { 0xD3, 0xF0, 0x16, 0x00, 0x76, 0x30, 0xAA };
	debugbreak = memory->find( signature, 13 );
	haveDebugBreak = debugbreak != 0xFFFFFFFF;
	if( haveDebugBreak ) {
		debugstr = ((memory->cread16(debugbreak - 4) << 16 ) | 
		             memory->cread16(debugbreak - 2)) & 0x003fffff  ;
		debugbreak -= 10;
		cout << "found @ " << hex << debugbreak << " : " << debugstr << dec<< endl;
	}
}
u16  CPUCore::r( int n ) {
	assert( n >= 0 && n <= 15 );
	return R[n];
}
void CPUCore::r( int n, u16 val ) {
	assert( n >= 0 && n <= 15 );
	R[n] = val;
}
u6   CPUCore::e( int n ) {
	assert( n >= 8 && n <= 15 );
	return (u6)(E[n-8] & 0x3f);
}
void CPUCore::e( int n, u6 val ) {
	assert( n >= 8 && n <= 15 );
	E[n-8] = (u6)(val & 0x3f);
}
u22  CPUCore::a( int n ) {
	assert( n >= 8 && n <= 15 );
	return ((u22)e(n) << 16) | (u22)r(n);
}
void CPUCore::a( int n, u22 val ) {
	assert( n >= 8 && n <= 15 );
	r( n, (u16)(val & 0x00ffff) );
	e( n, (u6)((val & 0x3f0000) >> 16) );
}
void CPUCore::a( int n, u6 vale, u16 valr ) {
	assert( n >= 8 && n <= 15 );
	e( n, vale );
	r( n, valr );
}
u16  CPUCore::d( int n ) {
	assert( n >= 0 && n <= 7 );
	return r(n);
}
void CPUCore::d( int n, u16 val ) {
	assert( n >= 0 && n <= 7 );
	r( n, val );
}
void CPUCore::dump() {
	dumpops(10);
	dumpregs();
}
void CPUCore::dumpregs() {
	updatedump();
	cout << dumpstring << endl;
}
void CPUCore::dumpops( u32 i ) {
	char opstring[256];
	if( i == 0 )
		i = perfops;
	assert( i <= OPHISTORY );
	if( i > steps )
		i = steps;
	i = (OPHISTORY + thisop - i + 1) % OPHISTORY;
	while( i != thisop ) {
		printop( i, opstring );
		cout << opstring;
		i = (i + 1) % OPHISTORY;
	}
	printop( i, opstring );
	cout << opstring;
}
int CPUCore::printop( u32 i, char* opstring ) {
	int size = opsize(ophistoryN[i]);
	if(  size == 2 )
		return sprintf( opstring, "%c%06X: %04x      (%03d) %s\n",
		   ophistorym[i], ophistoryPC[i], ophistory1[i], ophistoryN[i],
		   formatOP(ophistoryN[i], ophistory1[i], ophistory2[i], ophistoryPC[i] + size ));
	else
		return sprintf( opstring, "%c%06X: %04x %04x (%03d) %s\n",
		   ophistorym[i], ophistoryPC[i], ophistory1[i],  ophistory2[i], ophistoryN[i],
		   formatOP(ophistoryN[i], ophistory1[i], ophistory2[i], ophistoryPC[i] + size ));
}
void CPUCore::updateops( errcode erc, int mnnum, u16 op1, u16 op2 ) {
	thisop = (thisop + 1) % OPHISTORY;
	if( lastop == thisop )
		lastop = (lastop + 1) % OPHISTORY;
	switch( erc ) {
		case SUCCESS:
		case BREAKPOINT:
			++usage[mnnum];
			++steps;
			ophistorym[thisop] = '>';
			break;
		case ABSJUMP:
			++usage[mnnum];
			++steps;
			ophistorym[thisop] = '#';
			break;
		case RELJUMPUP:
			++usage[mnnum];
			++steps;
			ophistorym[thisop] = '^';
			break;
		case RELJUMPDN:
			++usage[mnnum];
			++steps;
			ophistorym[thisop] = 'v';
			break;
		case NOTIMPLEMENTED:
			ophistorym[thisop] = 'N';
			break;
		case INVALID:
			ophistorym[thisop] = 'I';
			break;
		case RESERVED:
			ophistorym[thisop] = 'R';
			break;
		case OPSIZE:
			ophistorym[thisop] = 'S';
			break;
		case MEMORY:
			ophistorym[thisop] = 'M';
			break;
		case READONLY:
			ophistorym[thisop] = '-';
			break;
		case WRITEONLY:
			ophistorym[thisop] = '+';
			break;
		case KNOWNDATA:
			ophistorym[thisop] = 'D';
			break;
		default:
			ophistorym[thisop] = '?';
	}
	ophistoryN[thisop] = mnnum;
	ophistoryPC[thisop] = oldPC;
	ophistory1[thisop]  = op1;
	ophistory2[thisop]  = op2;
}
void CPUCore::updatedump() {
	u16 stackval[20];
	u22 stackptr = a(15);
	int j = 0;
	if( stackptr >= 0x0fffff )
		for( int i = 2; i <= 40 && ((stackptr + i) < 0x200000); i+=2 )
			stackval[j++] = memory->dread16( a(15) + i );
	while( j < 20 )
		stackval[j++] = 0;
	sprintf( dumpstring,
	   "Steps: %10d"      "                      SP+ 2: %04x SP+22: %04x\n"
	   "R0= %04x R8 = %04x E8 = %02x A8 = %06x"" SP+ 4: %04x SP+24: %04x\n"
	   "R1= %04x R9 = %04x E9 = %02x A9 = %06x"" SP+ 6: %04x SP+26: %04x\n"
	   "R2= %04x R10= %04x E10= %02x A10= %06x"" SP+ 8: %04x SP+28: %04x\n"
	   "R3= %04x R11= %04x E11= %02x A11= %06x"" SP+10: %04x SP+30: %04x\n"
	   "R4= %04x R12= %04x E12= %02x A12= %06x"" SP+12: %04x SP+32: %04x\n"
	   "R5= %04x R13= %04x E13= %02x A13= %06x"" SP+14: %04x SP+34: %04x\n"
	   "R6= %04x R14= %04x E14= %02x A14= %06x"" SP+16: %04x SP+36: %04x\n"
	   "R7= %04x R15= %04x E15= %02x A15= %06x"" SP+18: %04x SP+38: %04x\n"
	   "T:%d PM:%d Z1:%d Z0:%d V:%d TE:%d IE:%d FE:%d  SP+20: %04x SP+40: %04x",
	   steps, stackval[0], stackval[10],
	   oldR[0], oldR[ 8], oldE[0], (oldE[0] << 16) | oldR[ 8], stackval[1], stackval[11],
	   oldR[1], oldR[ 9], oldE[1], (oldE[1] << 16) | oldR[ 9], stackval[2], stackval[12],
	   oldR[2], oldR[10], oldE[2], (oldE[2] << 16) | oldR[10], stackval[3], stackval[13],
	   oldR[3], oldR[11], oldE[3], (oldE[3] << 16) | oldR[11], stackval[4], stackval[14],
	   oldR[4], oldR[12], oldE[4], (oldE[4] << 16) | oldR[12], stackval[5], stackval[15],
	   oldR[5], oldR[13], oldE[5], (oldE[5] << 16) | oldR[13], stackval[6], stackval[16],
	   oldR[6], oldR[14], oldE[6], (oldE[6] << 16) | oldR[14], stackval[7], stackval[17],
	   oldR[7], oldR[15], oldE[7], (oldE[7] << 16) | oldR[15], stackval[8], stackval[18],
	   flagT(), flagPM(), flagZ1(), flagZ0(),
	   flagV(), flagTE(), flagIE(), flagFE(), stackval[9], stackval[19] );
}
void CPUCore::stats() {
	for( int i = 0; i < 283; ++i )
		cout << i << ":" << usage[i] << endl;
}
void CPUCore::go( u22 address ) {
	oldPC = PCC = address;
}
void CPUCore::walk( u22 address ) {
	errcode err = SUCCESS;
	while( (PCC != address) && (err = step()) == SUCCESS ) {
		dumpregs();
		dumpops(perfops);
	}
	stopping( (err != BREAKPOINT) ? (((PCC == address) ? REACHED : ERROR)) : BREAK, "walking" );
}
void CPUCore::jog( u22 address ) {
	errcode err = SUCCESS;
	while( (PCC != address) && (err = step()) == SUCCESS )
		dumpops(perfops);
	stopping( (err != BREAKPOINT) ? (((PCC == address) ? REACHED : ERROR)) : BREAK, "jogging" );
}
void CPUCore::run( u22 address ) {
	errcode err = SUCCESS;
	while( (PCC != address) && (err = step()) == SUCCESS )
		;
	stopping( (err != BREAKPOINT) ? (((PCC == address) ? REACHED : ERROR)) : BREAK, "running" );
}
void CPUCore::stopping( reason_t why, char *what ) {
	cout << "Stopped " << what << " at " << hex << PCC;
	switch( why ) {
		case REACHED:
			cout << " target address reached" << endl;
			break;
		case ERROR:
			cout << " due to error" << endl;
			break;
		case BREAK:
			cout << " breakpoint reached" << endl;
			break;
	}
	dumpops(perfops);
	dumpregs();
}
CPUCore::errcode CPUCore::step() {
	errcode result = doStep();
	switch( result ) {
		case SUCCESS:		
		case ABSJUMP:
		case RELJUMPUP:
		case RELJUMPDN:
			verify();
			return SUCCESS;
		case BREAKPOINT:
			verify();
			BL.Print( BL.getBreak() );
			return BREAKPOINT;
		case NOTIMPLEMENTED:
		case INVALID:
		case RESERVED:
		case OPSIZE:
		case MEMORY:
		case READONLY:
		case WRITEONLY:
		case KNOWNDATA:
		case OOPS:
			cout << "opcode returned: " << errcodestr[result] << endl;
			return OOPS;
		default:
			cout << "opcode returned invalid error code" << endl;
			return OOPS;
	}
}
bool CPUCore::verify() {
	doOveride();
	oldR[ 0]   = R[ 0];
	oldR[ 1]   = R[ 1];
	oldR[ 2]   = R[ 2];
	oldR[ 3]   = R[ 3];
	oldR[ 4]   = R[ 4];
	oldR[ 5]   = R[ 5];
	oldR[ 6]   = R[ 6];
	oldR[ 7]   = R[ 7];
	oldR[ 8]   = R[ 8];
	oldR[ 9]   = R[ 9];
	oldR[10]   = R[10];
	oldR[11]   = R[11];
	oldR[12]   = R[12];
	oldR[13]   = R[13];
	oldR[14]   = R[14];
	oldR[15]   = R[15];
	oldE[ 0]   = E[0];
	oldE[ 1]   = E[1];
	oldE[ 2]   = E[2];
	oldE[ 3]   = E[3];
	oldE[ 4]   = E[4];
	oldE[ 5]   = E[5];
	oldE[ 6]   = E[6];
	oldE[ 7]   = E[7];
	oldPC      = PCC;
	oldSPC_FIQ = SPC_FIQ;
	oldSPC_IRQ = SPC_IRQ;
	oldSR      = SR;
	oldSPSR[0] = SPSR[0];
	oldSPSR[1] = SPSR[1];
	oldSPSR[2] = SPSR[2];
	oldSPSR[3] = SPSR[3];
	oldSPSR[4] = SPSR[4];
	oldSPSR[5] = SPSR[5];
	oldSPSR[6] = SPSR[6];
	return true;
}
CPUCore::errcode CPUCore::doStep() {
	if( PCC >= MMU::MEMFIQLOCKED )
		return MEMORY;
	perfops = 1;
	u16 op = getOP();
	u16 op2 = 0;
	int mnnum = decodeOP( op );
	int size = opsize( mnnum );
	errcode result = OOPS;
	if( memory->isData( PCC ) ) { // - 1
		ungetOP();
		result = KNOWNDATA;
	}
	else if( size == 2 )
		result = doop( mnnum, op );
	else
		result = doop( mnnum, op, op2 = getOP(true) );
	updateops( result, mnnum, op, op2 );
	if( BL.isBreakpoint( oldPC, Breakpoint::X ))
		return BREAKPOINT;
	return result;
}
CPUCore::errcode CPUCore::delaySlot() {
	++perfops;
	u16 op = getOP();
	int mnnum = decodeOP( op );
	int size = opsize( mnnum );
	errcode result = OOPS;
	if( memory->isData( PCC ) ) { // - 1
		ungetOP();
		result = KNOWNDATA;
	}
	else if( size != 2 ) {
		ungetOP();
		result = OPSIZE;
	}
	else
		result = doop( mnnum, op );
	updateops( result, mnnum, op );
	return result;
}
void CPUCore::doOveride() {
	if( haveDebugBreak && PCC == debugbreak ) {
		cout << "debugstring" << endl;
		memory->prncstr( debugstr, 128 );
	}
	switch( version ) {
		case gm120pre:
			break;
		case gm120v120:
			break;
		case gm120v130:
			break;
		case gm120v140:
			break;
		case gm120v150:
			break;
		case gm120v170:
			break;
		case gm120v180:
			break;
		case gm120v190:
			break;
		case gm120v193:
			break;
		case gm120v1100:
			break;
		case gm220v160:
			break;
		case gm220v170:
			switch( PCC ) {
				case 0x6146:
				case 0x467B8:
					overide( ovFlagT, 1 );
					break;
				case 0x615A:
					overide( ovFlagZ0, 1 );
					break;
			}
			break;
		case gm220v180:
			break;
		case gmSPv110:
			break;
		case gmSPv120:
			break;
		case gmSPv123:
			break;
		case gmSPv130:
			break;
	}
}
void CPUCore::overide( Overides what, u32 value ) {
	switch( what ) {
		case ovFlagT:
			flagT( value );
			break;
		case ovFlagPM:
			flagPM( value );
			break;
		case ovFlagZ1:
			flagZ1( value );
			break;
		case ovFlagZ0:
			flagZ0( value );
			break;
		case ovFlagV:
			flagV( value );
			break;
		case ovFlagTE:
			flagTE( value );
			break;
		case ovFlagIE:
			flagIE( value );
			break;
		case ovFlagFE:
			flagFE( value );
			break;
		case ovR0:
		case ovR1:
		case ovR2:
		case ovR3:
		case ovR4:
		case ovR5:
		case ovR6:
		case ovR7:
		case ovR8:
		case ovR9:
		case ovR10:
		case ovR11:
		case ovR12:
		case ovR13:
		case ovR14:
		case ovR15:
			r( what - ovFlagFE, value );
		break;
		case ovE8:
		case ovE9:
		case ovE10:
		case ovE11:
		case ovE12:
		case ovE13:
		case ovE14:
		case ovE15:
			e( what - ovR15 + 8, value );
		break;
		case ovA8:
		case ovA9:
		case ovA10:
		case ovA11:
		case ovA12:
		case ovA13:
		case ovA14:
		case ovA15:
			a( what - ovE15 + 8, value );
	}
}

CPUCore::errcode CPUCore::branch( s22 off, bool save, bool delay ) {
	errcode result = SUCCESS;
	if( delay ) {
		result = delaySlot();
		ungetOP();
	}
	if( save )
		a( 14, PCC );
	PCC += off;
	if( memory->isData( PCC )) // - 1
		return MEMORY;
	memory->addmap( PCC, MMU::RJUMP );
	if(  result == SUCCESS ) {
		if( off < 0 )
			result = RELJUMPUP;
		else if( off > 0 )
			result = RELJUMPDN;
	}
	return result;
}
CPUCore::errcode CPUCore::jump( u22 addr, bool save, bool delay ) {
	errcode result = SUCCESS;
	if( delay )
		result = delaySlot();
	if( save )
		a( 14, PCC );
	PCC = addr;
	if( memory->isData( PCC )) // - 1
		return MEMORY;
	memory->addmap( addr, MMU::AJUMP );
	if(  result == SUCCESS )
		result = ABSJUMP;
	return result;
}
u16 CPUCore::getOP( bool op32 ) {
	u16 op;
	op = memory->cread( PCC, op32 );
	PCC += 2;
	return op;
}
void CPUCore::ungetOP() {
	PCC -= 2;
}
CPUCore::errcode CPUCore::doop( int mnnum, u16 op1, u16 op2 ) {
	switch( mnnum ) {
		case   0: return opADD2( op1 );
		case   1: return opSUB2( op1 );
		case   2: return opLD3( op1 );
		case   3: return opLDW1( op1 );
		case   4: return opLDW1( op1 );
		case   5: return opLDW2( op1 );
		case   6: return opLDW4( op1 );
		case   7: return opLDW2( op1 );
		case   8: return opLDW4( op1 );
		case   9: return opLDB1( op1 );
		case  10: return opLDB3( op1 );
		case  11: return opLDW5( op1 );
		case  12: return opLDW7( op1 );
		case  13: return opLDB1( op1 );
		case  14: return opLDB3( op1 );
		case  15: return opLDW5( op1 );
		case  16: return opLDW7( op1 );
		case  17: return opADD1( op1 );
		case  18: return opSUB1( op1 );
		case  19: return opADC1( op1 );
		case  20: return opSBC1( op1 );
		case  21: return opAND1( op1 );
		case  22: return opOR1( op1 );
		case  23: return opXOR1( op1 );
		case  24: return opTST1( op1 );
		case  25: return opCMP1( op1 );
		case  26: return opCMP1( op1 );
		case  27: return opCMP1( op1 );
		case  28: return opCMP1( op1 );
		case  29: return opCMP1( op1 );
		case  30: return opLD1( op1 );
		case  31: return opRR( op1 );
		case  32: return opRL( op1 );
		case  33: return opRRC( op1 );
		case  34: return opSRB( op1 );
		case  35: return opSR( op1 );
		case  36: return opSRA( op1 );
		case  37: return opJMP1( op1 );
		case  38: return opJMP1( op1 );
		case  39: return opJMP1( op1 );
		case  40: return opJMP1( op1 );
		case  41: return opSLB( op1 );
		case  42: return opDT( op1 );
		case  43: return opINCC( op1 );
		case  44: return opDECC( op1 );
		case  45: return opCOM( op1 );
		case  46: return opCOM( op1 );
		case  47: return opCOM( op1 );
		case  48: return opEXT( op1 );
		case  49: return opADD3( op1, op2 );
		case  50: return opADD5( op1, op2 );
		case  51: return opSUB4( op1, op2 );
		case  52: return opADC2( op1, op2 );
		case  53: return opSBC2( op1, op2 );
		case  54: return opAND3( op1, op2 );
		case  55: return opOR3( op1, op2 );
		case  56: return opXOR3( op1, op2 );
		case  57: return opTST3( op1, op2 );
		case  58: return opCMP2( op1, op2 );
		case  59: return opCMP2( op1, op2 );
		case  60: return opCMP2( op1, op2 );
		case  61: return opCMP2( op1, op2 );
		case  62: return opCMP2( op1, op2 );
		case  63: return opLD4( op1, op2 );
		case  64: return opReserved( op1 );
		case  65: return opCMPEQ2( op1 );
		case  66: return opAND2( op1 );
		case  67: return opOR2( op1 );
		case  68: return opXOR2( op1 );
		case  69: return opTST2( op1 );
		case  70: return opLDB4( op1 );
		case  71: return opLDB4( op1 );
		case  72: return opBIT( op1 );
		case  73: return opBIT( op1 );
		case  74: return opBIT( op1 );
		case  75: return opBIT( op1 );
		case  76: return opSYS( op1 );
		case  77: return opSWI( op1 );
		case  78: return opCLRSR( op1 );
		case  79: return opSETSR( op1 );
		case  80: return opTSTSR( op1 );
		case  81: return opNOP( op1 );
		case  82: return opBREAK( op1 );
		case  83: return opLDSR( op1 );
		case  84: return opLDSR( op1 );
		case  85: return opRET_FIQ( op1 );
		case  86: return opRET_IRQ( op1 );
		case  87: return opRET_SWI( op1 );
		case  88: return opRETD( op1 );
		case  89: return opLDSvR1( op1 );
		case  90: return opLDSvR1( op1 );
		case  91: return opLDSvR1( op1 );
		case  92: return opReserved( op1 );
		case  93: return opLDSvR1( op1 );
		case  94: return opLDSvR1( op1 );
		case  95: return opLDSvR1( op1 );
		case  96: return opReserved( op1 );
		case  97: return opReserved( op1 );
		case  98: return opLDSvR1( op1 );
		case  99: return opReserved( op1 );
		case 100: return opReserved( op1 );
		case 101: return opLDSvR2( op1 );
		case 102: return opLDSvR2( op1 );
		case 103: return opLDSvR2( op1 );
		case 104: return opReserved( op1 );
		case 105: return opLDSvR2( op1 );
		case 106: return opLDSvR2( op1 );
		case 107: return opLDSvR2( op1 );
		case 108: return opReserved( op1 );
		case 109: return opReserved( op1 );
		case 110: return opLDSvR2( op1 );
		case 111: return opReserved( op1 );
		case 112: return opReserved( op1 );
		case 113: return opReserved( op1 );
		case 114: return opReserved( op1 );
		case 115: return opLDPC( op1 );
		case 116: return opReserved( op1 );
		case 117: return opJMP2( op1, op2 );
		case 118: return opJMP2( op1, op2 );
		case 119: return opJMP2( op1, op2 );
		case 120: return opJMP2( op1, op2 );
		case 121: return opLDC( op1 );
		case 122: return opReserved( op1 );
		case 123: return opLDRE( op1 );
		case 124: return opLDRE( op1 );
		case 125: return opCMPEQ1( op1 );
		case 126: return opLD2( op1 );
		case 127: return opLDW3( op1, op2 );
		case 128: return opLDW3( op1, op2 );
		case 129: return opLDB2( op1, op2 );
		case 130: return opLDB2( op1, op2 );
		case 131: return opLDW6( op1, op2 );
		case 132: return opLDW6( op1, op2 );
		case 133: return opCMP3( op1 );
		case 134: return opADD6( op1 );
		case 135: return opSUB5( op1 );
		case 136: return opCMPEQ3( op1, op2 );
		case 137: return opLD5( op1, op2 );
		case 138: return opADD4( op1 );
		case 139: return opSUB3( op1 );
		case 140: return opMUL( op1 );
		case 141: return opMUL( op1 );
		case 142: return opMUL( op1 );
		case 143: return opMUL( op1 );
		case 144: return opPOP1( op1 );
		case 145: return opReserved( op1 );
		case 146: return opPOP2( op1 );
		case 147: return opPUSH1( op1 );
		case 148: return opReserved( op1 );
		case 149: return opPUSH2( op1 );
		case 150: return opBSRD( op1 );
		case 151: return opBRAEC( op1 );
		case 152: return opReserved( op1 );
		case 153: return opBRAEC( op1 );
		case 154: return opBNZD( op1 );
		case 155: return opReserved( op1 );
		case 156: return opBR( op1 );
		case 157: return opBR( op1 );
		case 158: return opBR( op1 );
		case 159: return opBR( op1 );
		case 160: return opBR( op1 );
		case 161: return opBR( op1 );
		case 162: return opCLD( op1 );
		case 163: return opCLD( op1 );
		case 164: return opECLD( op1 );
		case 165: return opELD5( op1, op2 );
		case 166:
			switch((op1 & 0x0060) >> 5) {
				case 0: return opELD1( op1, op2 );
				case 1: return opEADD1( op1, op2 );
				case 2: return opESUB1( op1, op2 );
				case 3: return opECP1( op1, op2 );
			}
		case 167: return opELD5( op1, op2 );
		case 168: return opERPN( op1, op2 );
		case 169: return opELD1( op1, op2 );
		case 170: return opELD5( op1, op2 );
		case 171: return opEMAD2( op1 );
		case 172: return opEMSB2( op1 );
		case 173: return opEMLD2( op1 );
		case 174: return opEMUL2( op1 );
		case 175: return opEADD4( op1 );
		case 176: return opESUB4( op1 );
		case 177: return opEADD4( op1 );
		case 178: return opESUB4( op1 );
		case 179: return opELD4( op1 );
		case 180: return opELD4( op1 );
		case 181: return opEADD4( op1 );
		case 182: return opESUB4( op1 );
		case 183: return opELD4( op1 );
		case 184: return opEADD4( op1 );
		case 185: return opESUB4( op1 );
		case 186: return opELD4( op1 );
		case 187: return opReserved( op1 );
		case 188: return opEADD2( op1 );
		case 189: return opESUB2( op1 );
		case 190: return opELD2( op1 );
		case 191: return opEADD2( op1 );
		case 192: return opESUB2( op1 );
		case 193: return opELD2( op1 );
		case 194: return opELD5( op1 );
		case 195: return opReserved( op1 );
		case 196: return opEADD2( op1 );
		case 197: return opESUB2( op1 );
		case 198: return opELD2( op1 );
		case 199: return opEADD2( op1 );
		case 200: return opESUB2( op1 );
		case 201: return opELD2( op1 );
		case 202: return opELD5( op1 );
		case 203: return opReserved( op1 );
		case 204: return opEADD2( op1 );
		case 205: return opESUB2( op1 );
		case 206: return opECP2( op1 );
		case 207: return opEMAX2( op1 );
		case 208: return opEMIN2( op1 );
		case 209: return opELD5( op1 );
		case 210: return opELD1( op1 );
		case 211: return opELD5( op1 );
		case 212: return opELD5( op1 );
		case 213: return opELD5( op1 );
		case 214: return opELD5( op1 );
		case 215: return opEMAD3( op1 );
		case 216: return opEMSB3( op1 );
		case 217: return opEMLD3( op1 );
		case 218: return opEMUL3( op1 );
		case 219: return opELD6( op1 );
		case 220: return opReserved( op1 );
		case 221: return opReserved( op1 );
		case 222: return opESFT( op1 );
		case 223: return opESFTA( op1 );
		case 224: return opESFTL( op1 );
		case 225: return opESFTD( op1 );
		case 226: return opELD5( op1 );
		case 227: return opENMSK( op1 );
		case 228: return opELD5( op1 );
		case 229: return opELD5( op1 );
		case 230: return opELD5( op1 );
		case 231: return opESD( op1 );
		case 232: return opESD( op1 );
		case 233: return opESD( op1 );
		case 234: return opESD( op1 );
		case 235: return opELD1( op1 );
		case 236: return opELD5( op1 );
		case 237:
			switch((op1 & 0x0180) >> 7) {
				case 0: return opELD1( op1 );
				case 1: return opEADD1( op1 );
				case 2: return opESUB1( op1 );
				case 3: return opECP1( op1 );
			}
		case 238:
			switch((op1 & 0x000c) >> 2) {
				case 0: return opELD1( op1 );
				case 1: return opEADD1( op1 );
				case 2: return opESUB1( op1 );
				case 3: return opECP1( op1 );
			}
		case 239:
			switch((op1 & 0x000c) >> 2) {
				case 0: return opELD1( op1 );
				case 1: return opEADD1( op1 );
				case 2: return opESUB1( op1 );
				case 3: return opECP1( op1 );
			}
		case 240: return opELD5( op1 );
		case 241: return opEMAD2( op1 );
		case 242: return opEMSB2( op1 );
		case 243: return opEMLD2( op1 );
		case 244: return opEMUL2( op1 );
		case 245: return opEMAX1( op1 );
		case 246: return opEMIN1( op1 );
		case 247: return opEMAX1( op1 );
		case 248: return opEMIN1( op1 );
		case 249: return opENOP( op1 );
		case 250: return opELD5( op1 );
		case 251: return opELD5( op1 );
		case 252: return opEMAD2( op1 );
		case 253: return opEMSB2( op1 );
		case 254: return opEMLD2( op1 );
		case 255: return opEMUL2( op1 );
		case 256: return opEMAD1( op1 );
		case 257: return opEMSB1( op1 );
		case 258: return opEMAD2( op1 );
		case 259: return opEMSB2( op1 );
		case 260: return opEMLD2( op1 );
		case 261: return opEMUL2( op1 );
		case 262: return opEMLD1( op1 );
		case 263: return opEMUL1( op1 );
		case 264: return opERPR( op1 );
		case 265: return opELD1( op1 );
		case 266: return opEADD1( op1 );
		case 267: return opECP1( op1 );
		case 268:
			switch(op1 & 0x000f) {
				case 0: return opESRA1( op1 );
				case 1: return opESLA1( op1 );
				case 2: return opESRA8( op1 );
				case 3: return opESLA( op1 );
				case 4: return opESRC( op1 );
				case 5: return opESLC( op1 );
				case 6: return opEINCC( op1 );
				case 7: return opEDECC( op1 );
				case 8: return opENEG( op1 );
				case 9: return opEABS( op1 );
				case 10: return opEFS16( op1 );
				case 11: return opEFZ16( op1 );
				case 12: return opEFS8( op1 );
				case 13: return opEFZ8( op1 );
				case 14: return opEEXP( op1 );
				case 15: return opEEXPC( op1 );
			}
		case 269: return opERPN( op1 );
		case 270: return opERPS( op1 );
		case 271: return opERPD( op1 );
		case 272:
			switch(op1 & 0x000f) {
				case 0: return opESRA2( op1 );
				case 1: return opESLA2( op1 );
				case 2: return opERND( op1 );
				case 3: return opECR( op1 );
				case 4: return opESAT( op1 );
				case 5: return opERESR( op1 );
				case 6: return opReserved( op1 );
				case 7: return opReserved( op1 );
				case 8: return opELD3( op1 );
				case 9: return opReserved( op1 );
				case 10: return opEADD3( op1 );
				case 11: return opESUB3( op1 );
				case 12: return opEADD3( op1 );
				case 13: return opESUB3( op1 );
				case 14: return opEDIVQ( op1 );
				case 15: return opReserved( op1 );
			}
		case 273: return opETST( op1 );
		case 274: return opES( op1 );
		case 275: return opELD5( op1 );
		case 276: return opELD5( op1 );
		case 277: return opELD5( op1 );
		case 278: return opELD5( op1 );
		case 279: return opEBK( op1 );
		case 280: return opESEC( op1 );
		case 281: return opESEC( op1 );
		case 282: return opESEC( op1 );
		default: return opInvalid( op1 );
	}
}
int  CPUCore::subSPSR( SPSRegister n ) {
	switch( n ) {
		case SPCL_FIQ:
			return 0;
		case SPCH_FIQ:
			return 1;
		case SSR_FIQ:
			return 2;
		case SPCL_IRQ:
			return 3;
		case SPCH_IRQ:
			return 4;
		case SSR_IRQ:
			return 5;
		case SSR_SWI:
			return 6;
		default:
			assert( false );
	}
	return false;
}
u16 CPUCore::sub2scomp( u16 val ) {
	val ^= 0xffff;
	return ++val;
}
CPUCore::errcode CPUCore::subADSBC( int n, u16 v2, bool c, bool m ) {
	u16 v1 = r(n);
	if(m)
		v2 = (u16)(sub2scomp( v2 ) - (c?1:0));
	u32 val = v1 + v2;
	if(c & !m) {
		++val;
	}
	if(m) {
		flagT( !(val & 0x10000) && (c || v2) );
		flagV( ((v1 & 0x8000) != (v2 & 0x8000)) && ((u16)(v1 & (u16)0x8000) != (u16)(val & 0x8000)) );
	}
	else {
		flagT( val & 0x10000  );
		flagV( ((v1 & 0x8000) == (v2 & 0x8000)) && ((u16)(v1 & (u16)0x8000) != (u16)(val & 0x8000)) );
	}
	r( n, (u16)val );
	if( n == 6 )
		flagZ0( val == 0 );
	if( n == 7 )
		flagZ1( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::subAND( int n, u16 v2 ) {
	u16 v1 = r(n);
	u16 val = v1 & v2;
	flagT( val == 0 );
	r(n, val );
	if( n == 6 )
		flagZ0( val == 0 );
	if( n == 7 )
		flagZ1( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opADC1( u16 op1 ) {
	return subADSBC( (op1 & 0xf00) >> 8, r(op1 & 0xf), flagT(), false );
}
CPUCore::errcode CPUCore::opADC2( u16 op1, u16 op2 ) {
	return subADSBC( op1 & 0xf, op2, flagT(), false );
}
CPUCore::errcode CPUCore::opADD1( u16 op1 ) {
	return subADSBC( (op1 & 0xf00) >> 8, r(op1 & 0xf), false, false );
}
CPUCore::errcode CPUCore::opADD2( u16 op1 ) {
	return subADSBC( (op1 & 0xf00) >> 8, (u16)(op1 & 0x7f), false,false );
}
CPUCore::errcode CPUCore::opADD3( u16 op1, u16 op2 ) {
	return subADSBC( op1 & 0xf, op2, false, false );
}
CPUCore::errcode CPUCore::opADD4( u16 op1 ) {
	int n = ((op1 & 0x700) >> 8) + 8;
	int i = op1 & 0xf;
	a( n, a(n) + r(i) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opADD5( u16 op1, u16 op2 ) {
	int n = (op1 & 0x0007) + 8;
	a( n, a(n) + op2 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opADD6( u16 op1 ) {
	int n = ((op1 & 0x700) >> 8) + 8;
	a( n, a(n) + (op1 & 0x1f) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opAND1( u16 op1 ) {
	return subAND( (op1 & 0x0f00) >> 8, r(op1 & 0x000f) );
}
CPUCore::errcode CPUCore::opAND2( u16 op1 ) {
	return subAND( 0, (u16)(op1 & 0x00ff) );
}
CPUCore::errcode CPUCore::opAND3( u16 op1, u16 op2 ) {
	return subAND( op1 & 0x000f, op2 );
}
CPUCore::errcode CPUCore::opBIT( u16 op1 ) {
	u22 addr;
	u8 val, mask;
	addr = a(8) + r(1);
	if( BL.isBreakpoint( addr, Breakpoint::R ))
		return BREAKPOINT;
	if( ((op1 & 0x0018) >> 3) == 3) {
		if( BL.isBreakpoint( addr, Breakpoint::W ))
			return BREAKPOINT;
	}
	val = memory->dread8( addr );
	mask = (u8)(1 << (op1 & 0x0007));
	flagT( !(val & mask) );
	switch( (op1 & 0x0018) >> 3 ) {
		case 0:
			val &= (mask ^ (u8)0xff);
			break;
		case 1:
			val |= mask;
			break;
		case 2:
			val ^= mask;
			break;
		case 3:
			return SUCCESS;
	}
	return memory->dwrite8( addr, val )?SUCCESS:MEMORY;
}
CPUCore::errcode CPUCore::opBNZD( u16 op1 ) {
	int n = 7;
	if( !(op1 & 0x0100) )
		n = 6;
	u16 val = r(n);
	--val;
	r( n, val );
	s32 off = sext8to32((op1 & 0x007f) << 1);
	if( n == 6 && !flagZ0() )
		flagZ0( val == 0 );
	else if( n == 7 && !flagZ1() )
		flagZ1( val == 0 );
	else
		return SUCCESS;
	return branch( off, false, true );
}
CPUCore::errcode CPUCore::opBR( u16 op1 ) {
	bool delay = op1 & 0x0400;
	s32 off = sext11to32((op1 & 0x03ff) << 1);
	switch( (op1 & 0x1800) >> 11 ) {
		case 1: // BRA
			return branch( off, false, delay );
		case 2: // BRF
			if( !flagT() )
				return branch( off, false, delay );
			break;
		case 3: // BRT
			if( flagT() )
				return branch( off, false, delay );
			break;
		default:
			return INVALID;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opBRAEC( u16 op1 ) {
	if( EC[(op1 & 0x0180) >> 7] )
		return branch( sext8to32((op1 & 0x007f) << 1), false, false );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opBREAK( u16 op1 ) {
	return SUCCESS; // asserts nBRK signal for one cycle
}
CPUCore::errcode CPUCore::opBSRD( u16 op1 ) {
	++depth;
	return branch( sext11to32((op1 & 0x03ff) << 1), true, true );
}
CPUCore::errcode CPUCore::opCLD( u16 op1 ) {
	return NOTIMPLEMENTED;
}
CPUCore::errcode CPUCore::opCLRSR( u16 op1 ) {
	switch( op1 & 0x0007 ) {
		case 0:
			flagFE( false );
			break;
		case 1:
			flagIE( false );
			break;
		case 2:
			flagTE( false );
			break;
		case 3:
			flagV( false );
			break;
		case 4:
			flagZ0( false );
			break;
		case 5:
			flagZ1( false );
			break;
		case 6:
			flagPM( false );
			break;
		default:
			assert( false );
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCMP1( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int i = op1 & 0x000f;
	switch( (op1 & 0x0070) >> 4 ) {
		case 0: //GE
			flagT( (s16)r(n) >= (s16)r(i) );
			break;
		case 1: //GT
			flagT( (s16)r(n) > (s16)r(i) );
			break;
		case 2: //UGE
			flagT( r(n) >= r(i) );
			break;
		case 3: //UGT
			flagT( r(n) > r(i) );
			break;
		case 4: //EQ
			flagT( r(n) == r(i) );
			break;
		default:
			assert( false );
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCMP2( u16 op1, u16 op2 ) {
	int n = op1 & 0x000f;
	switch( (op1 & 0x0700) >> 8 ) {
		case 0: //GE
			flagT( (s16)r(n) >= (s16)op2 );
			break;
		case 1: //GT
			flagT( (s16)r(n) > (s16)op2 );
			break;
		case 2: //UGE
			flagT( r(n) >= op2 );
			break;
		case 3: //UGT
			flagT( r(n) > op2 );
			break;
		case 4: //EQ
			flagT( r(n) == op2 );
			break;
		default:
			assert( false );
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCMP3( u16 op1 ) {
	flagT( (s16)d((op1 & 0x0700) >> 8) >= (s16)(op1 & 0x003f) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCMPEQ1( u16 op1 ) {
	flagT( a(((op1 & 0x0700) >> 8) + 8 ) == a((op1 & 0x0007) + 8) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCMPEQ2( u16 op1 ) {
	flagT( d((op1 & 0x0700) >> 8) == (op1 & 0x00ff) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCMPEQ3( u16 op1, u16 op2 ) {
	flagT( a( ((op1 & 0x0700) >> 8) + 8) == (u22)(((op1 & 0x003f) << 16) | op2) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCOM( u16 op1 ) {
	int n = op1 & 0x000f;
	u16 val = r(n) ^ (u16)0xffff;
	switch( (op1 & 0x0300) >> 8 ) {
		case 0:
			r( n, val );
			flagT( r(n) == 0 );
			break;
		case 1:
			++val;
			r( n, val );
			flagT( val == 0 );
			break;
		case 2:
			if( flagT() )
				++val;
			r( n, val );
			flagT( val == 0 );
			break;
		default:
			assert( false );
	}
	if( n == 6 )
		flagZ0( val == 0 );
	if( n == 7 )
		flagZ1( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opCOP( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opDECC( u16 op1 ) {
	return subADSBC( op1 & 0x000f, 1, flagT(), true );
}
CPUCore::errcode CPUCore::opDT( u16 op1 ) {
	return subADSBC( op1 & 0x000f, 1, false, true );
}
CPUCore::errcode CPUCore::opEXT( u16 op1 ) {
	int n = op1 & 0x000f;
	s16 val = sext8to16(r(n));
	r( n, val );
	if( n == 6 )
		flagZ0( val == 0 );
	if( n == 7 )
		flagZ1( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opINCC( u16 op1 ) {
	return subADSBC( op1 & 0x000f, 1, flagT(), false );
}
CPUCore::errcode CPUCore::opJMP1( u16 op1 ) {
	int i = (op1 & 0x0007) + 8;

	switch( ((op1 & 0x0100) >> 7) | ((op1 & 0x0008) >> 3) ) {
		case 0: //JPF
			if( !flagT() )
				return jump(a(i));
			break;
		case 1: //JPT
			if( flagT() )
				return jump(a(i));
			break;
		case 2: //JMP
			if( i == 14 && depth > 0)
				--depth;
			return jump(a(i));
		case 3: //JSR
			++depth;
			return jump(a(i), true);
		default:
			return INVALID;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opJMP2( u16 op1, u16 op2 ) {
	u22 imm = ((op1 & 0x0037) << 16) | op2;
	//imm += memory->imagebase();
	switch( (op1 & 0x00c0) >> 6 ) {
		case 0: //JPF
			if( !flagT() )
				return jump(imm);
			break;
		case 1: //JPT
			if( flagT() )
				return jump(imm);
			break;
		case 2: //JMP
			return jump(imm);
		case 3: //JSR
			++depth;
			return jump(imm, true);
		default:
			return INVALID;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLD1( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int i = (op1 & 0x000f);
	u16 val = r(i);
	r( n, val );
	if( n == 6 )
		flagZ0( val == 0 );
	if( n == 7 )
		flagZ1( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLD2( u16 op1 ) {
	a( ((op1 & 0x0700) >> 8) + 8, a((op1 & 0x0007) + 8) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLD3( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	u16 val = (u16)(op1 & 0x00ff);
	r( n, val );
	if( n == 6 )
		flagZ0( val == 0 );
	if( n == 7 )
		flagZ1( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLD4( u16 op1, u16 op2 ) {
	int n = op1 & 0x000f;
	r( n, op2);
	if( n == 6 )
		flagZ0( op2 == 0 );
	if( n == 7 )
		flagZ1( op2 == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLD5( u16 op1, u16 op2 ) {
	a( ((op1 & 0x0700) >> 8) + 8, (u6)(op1 & 0x003f), op2 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDRE( u16 op1 ) {
	int n = (op1 & 0x0700) >> 8;
	int i = (op1 & 0x0003) + 8;
	if(!(op1 & 0x0008))
		d(n, (u16)e(i));
	else
		e(i, (u6)(d(n) & 0x003f));
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB1( u16 op1 ) {
	int n = (op1 & 0x0700) >> 8;
	int i = ((op1 & 0x0070) >> 4) + 8;
	u22 disp = (op1 & 0x000f);
	if(!(op1 & 0x1000)) {
		if( BL.isBreakpoint( a(i) + disp, Breakpoint::R ))
			return BREAKPOINT;
		d(n, memory->dread8( a(i) + disp ) );
	} else {
		if( BL.isBreakpoint( a(i) + disp, Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite8( a(i) + disp, d(n) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB2( u16 op1, u16 op2 ) {
	int n = (op1 & 0x0700) >> 8;
	int i = (op1 & 0x0007) + 8;
	u22 disp = op2;
	if(!(op1 & 0x0008)) {
		if( BL.isBreakpoint( a(i) + disp, Breakpoint::R ))
			return BREAKPOINT;
		d(n, memory->dread8( a(i) + disp ) );
	} else {
		if( BL.isBreakpoint( a(i) + disp,  Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite8( a(i) + disp, d(n) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB3( u16 op1 ) {
	int n = (op1 & 0x0700) >> 8;
	int i = ((op1 & 0x0070) >> 4) + 8;
	int j = (op1 & 0x000f);
	if(!(op1 & 0x1000)) {
		if( BL.isBreakpoint( a(i) + r(j), Breakpoint::R ))
			return BREAKPOINT;
		d(n, memory->dread8( a(i) + r(j) ) );
	} else {
		if( BL.isBreakpoint( a(i) + r(j), Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite8( a(i) + r(j), d(n) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDB4( u16 op1 ) {
	u22 disp = (op1 & 0x00ff);
	if(!(op1 & 0x00100)) {
		if( BL.isBreakpoint( a(8) + disp, Breakpoint::R ))
			return BREAKPOINT;
		r(0, memory->dread8( a(8) + disp ) );
	} else {
		if( BL.isBreakpoint( a(8) + disp, Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite8( a(8) + disp, r(0) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDC( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int i = (op1 & 0x0007) + 8;
	if( BL.isBreakpoint( a(i), Breakpoint::R ))
		return BREAKPOINT;
	r(n, memory->cread16(a(i)));
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDPC( u16 op1 ) {
	int n = (op1 & 0x0007) + 8;
	a( n, PCC );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDSvR1( u16 op1 ) {
	r( 0, SPSR[op1 & 0x000f] );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDSvR2( u16 op1 ) {
	SPSR[op1 & 0x000f] = r(0);
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDSR( u16 op1 ) {
	if( op1 & 0x0001 )
		r( 0, SR );
	else
		r( 0, SR );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW1( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	u22 disp = ((op1 & 0x00ff) << 1);
	if(!(op1 & 0x1000)) {
		if( BL.isBreakpoint( a(15) + disp, Breakpoint::R ))
			return BREAKPOINT;
		r(n, memory->dread16( a(15) + disp ) );
	} else {
		if( BL.isBreakpoint( a(15) + disp, Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite16( a(15) + disp, r(n) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW2( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int m = ((op1 & 0x0070) >> 4) + 8;
	u22 disp = ((op1 & 0x000f) << 1);
	if(!(op1 & 0x1000)) {
		if( BL.isBreakpoint( a(m) + disp, Breakpoint::R ))
			return BREAKPOINT;
		r(n, memory->dread16( a(m) + disp ) );
	} else {
		if( BL.isBreakpoint( a(m) + disp, Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite16( a(m) + disp, r(n) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW3( u16 op1, u16 op2 ) {
	int n = (op1 & 0x0f00) >> 8;
	int m = (op1 & 0x0007) + 8;
	u22 disp = op2;
	if(!(op1 & 0x0008)) {
		if( BL.isBreakpoint( a(m) + disp, Breakpoint::R ))
			return BREAKPOINT;
		r(n, memory->dread16( a(m) + disp ) );
	} else {
		if( BL.isBreakpoint( a(m) + disp, Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite16( a(m) + disp, r(n) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW4( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int m = ((op1 & 0x0070) >> 4) + 8;
	int i = (op1 & 0x000f);
	if(!(op1 & 0x1000)) {
		if( BL.isBreakpoint( a(m) + r(i), Breakpoint::R ))
			return BREAKPOINT;
		r(n, memory->dread16( a(m) + r(i) ));
	} else {
		if( BL.isBreakpoint( a(m) + r(i), Breakpoint::W ))
			return BREAKPOINT;
		return memory->dwrite16( a(m) + r(i), r(n) )?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW5( u16 op1 ) {
	int n = ((op1 & 0x0700) >> 8) + 8;
	int i = ((op1 & 0x0070) >> 4) + 8;
	u22 disp = ((op1 & 0x000f) << 1);
	if(!(op1 & 0x1000)) {
		if( BL.isBreakpoint( a(i) + disp, Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( a(i) + disp + 2, Breakpoint::R ))
			return BREAKPOINT;
		a(n, memory->dread16( a(i) + disp ), memory->dread16( a(i) + disp + 2));
	} else {
		bool ok = true;
		if( BL.isBreakpoint( a(i) + disp, Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( a(i) + disp + 2, Breakpoint::W ))
			return BREAKPOINT;
		ok &= memory->dwrite16( a(i) + disp, e(n) );
		ok &= memory->dwrite16( a(i) + disp + 2, r(n) );
		return ok?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW6( u16 op1, u16 op2 ) {
	int n = ((op1 & 0x0700) >> 8) + 8;
	int i = (op1 & 0x0007) + 8;
	u22 disp = op2;
	if(!(op1 & 0x0008)) {
		if( BL.isBreakpoint( a(i) + disp, Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( a(i) + disp + 2, Breakpoint::R ))
			return BREAKPOINT;
		a(n, memory->dread16( a(i) + disp ), memory->dread16( a(i) + disp + 2));
	} else {
		bool ok = true;
		if( BL.isBreakpoint( a(i) + disp, Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( a(i) + disp + 2, Breakpoint::W ))
			return BREAKPOINT;
		ok &= memory->dwrite16( a(i) + disp, e(n) );
		ok &= memory->dwrite16( a(i) + disp + 2, r(n) );
		return ok?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opLDW7( u16 op1 ) {
	int n = ((op1 & 0x0700) >> 8) + 8;
	int i = ((op1 & 0x0070) >> 4) + 8;
	int j = (op1 & 0x000f);
	if(!(op1 & 0x1000)) {
		if( BL.isBreakpoint( a(i) + r(j), Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( a(i) + r(j) + 2, Breakpoint::R ))
			return BREAKPOINT;
		a(n, memory->dread16(a(i) + r(j)), memory->dread16(a(i) + r(j) + 2));
	} else {
		bool ok = true;
		if( BL.isBreakpoint( a(i) + r(j), Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( a(i) + r(j) + 2, Breakpoint::W ))
			return BREAKPOINT;
		ok &= memory->dwrite16( a(i) + r(j), e(n) );
		ok &= memory->dwrite16( a(i) + r(j) + 2, r(n) );
		return ok?SUCCESS:MEMORY;
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opMUL( u16 op1 ) {
	int n = ((op1 & 0x0700) >> 8);
	int i = (op1 & 0x0007);
	u8 va = (u8)(d(n) & 0xff);
//   bool sa = va & 0x80;
	u8 vb = (u8)(d(i) & 0xff);
//   bool sb = vb & 0x80;
	u16 val;
	switch( ((op1 & 0x0800) >> 10) | ((op1 & 0x0008) >> 3) ) {
		case 0: //UU
			val = (u16)(va * vb);
			break;
		case 1: //US
			val = (u16)(va * (s8)vb);
			break;
		case 2: //SU
			val = (u16)((s8)va * vb);
			break;
		case 3: //SS
			val = (u16)((s8)va * (s8)vb);
			break;
		default:
			return INVALID;
	}
	d( n, val );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opNOP( u16 op1 ) {return SUCCESS;}
CPUCore::errcode CPUCore::opOR1( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int i = op1 & 0x000f;
	u16 val = r(n) | r(i);
	bool t = val == 0;
	flagT(t);
	r(n, val );
	if( n == 6 )
		flagZ0(t);
	if( n == 7 )
		flagZ1(t);
	return SUCCESS;
}
CPUCore::errcode CPUCore::opOR2( u16 op1 ) {
	u16 val = r(0) | (u16)(op1 & 0x00ff);
	flagT( val == 0 );
	r(0, val );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opOR3( u16 op1, u16 op2 ) {
	int n = op1 & 0x000f;
	u16 val = r(op1 & 0x000f) | op2;
	bool t = val == 0;
	flagT( t );
	r( n, val );
	if( n == 6 )
		flagZ0(t);
	if( n == 7 )
		flagZ1(t);
	return SUCCESS;
}
CPUCore::errcode CPUCore::opPOP1( u16 op1 ) {
	int m = (op1 & 0x0f00) >> 8;
	int n = (op1 & 0x0007) | ((op1 & 0x0800) >> 8);
	assert( m != 15 && n != 15 );
	u22 addr = a(15);
	if( m == n ) {
		if( BL.isBreakpoint( addr + 2, Breakpoint::R ))
			return BREAKPOINT;
		r(n, memory->dread16(addr + 2));
		a(15, addr + 2);
	}
	else {
		if( BL.isBreakpoint( addr + 2, Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr + 4, Breakpoint::R ))
			return BREAKPOINT;
		r(n, memory->dread16(addr + 2));
		r(m, memory->dread16(addr + 4));
		a(15, addr + 4);
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opPOP2( u16 op1 ) {
	int m = ((op1 & 0x0700) >> 8) + 8;
	int n = (op1 & 0x0007) + 8;
	assert( m != 15 && n != 15 );
	u22 addr = a(15);
	if( m == n ) {
		if( BL.isBreakpoint( addr + 2, Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr + 4, Breakpoint::R ))
			return BREAKPOINT;
		e(n, (u6)(memory->dread16(addr + 2) & 0x003f));
		r(n, memory->dread16(addr + 4));
		a(15, addr + 4);
	}
	else {
		if( BL.isBreakpoint( addr + 2, Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr + 4, Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr + 6, Breakpoint::R ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr + 8, Breakpoint::R ))
			return BREAKPOINT;
		e(n, (u6)(memory->dread16(addr + 2) & 0x003f));
		r(n, memory->dread16(addr + 4));
		e(m, (u6)(memory->dread16(addr + 6) & 0x003f));
		r(m, memory->dread16(addr + 8));
		a(15, addr + 8);
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opPUSH1( u16 op1 ) {
	int m = (op1 & 0x0f00) >> 8;
	int n = (op1 & 0x0007) | ((op1 & 0x0800) >> 8);
	assert( m != 15 && n != 15 );
	u22 addr = a(15);
	bool ok = true;
	if( m == n ) {
		if( BL.isBreakpoint( addr, Breakpoint::W ))
			return BREAKPOINT;
		ok &= memory->dwrite16(addr, r(n));
		a(15, addr - 2);
	}
	else {
		if( BL.isBreakpoint( addr, Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr - 2, Breakpoint::W ))
			return BREAKPOINT;
		ok &= memory->dwrite16(addr, r(n));
		ok &= memory->dwrite16(addr - 2, r(m));
		a(15, addr - 4);
	}
	return ok?SUCCESS:MEMORY;
}
CPUCore::errcode CPUCore::opPUSH2( u16 op1 ) {
	int m = ((op1 & 0x0700) >> 8) + 8;
	int n = (op1 & 0x0007) + 8;
	assert( m != 15 && n != 15 );
	u22 addr = a(15);
	bool ok = true;
	if( m == n ) {
		if( BL.isBreakpoint( addr, Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr - 2, Breakpoint::W ))
			return BREAKPOINT;
		ok &= memory->dwrite16(addr, r(n));
		ok &= memory->dwrite16(addr - 2, e(n));
		a(15, addr - 4);
	}
	else {
		if( BL.isBreakpoint( addr, Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr - 2, Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr - 4, Breakpoint::W ))
			return BREAKPOINT;
		if( BL.isBreakpoint( addr - 6, Breakpoint::W ))
			return BREAKPOINT;
		ok &= memory->dwrite16(addr, r(n));
		ok &= memory->dwrite16(addr - 2, e(n));
		ok &= memory->dwrite16(addr - 4, r(m));
		ok &= memory->dwrite16(addr - 6, e(m));
		a(15, addr - 8);
	}
	return ok?SUCCESS:MEMORY;
}
CPUCore::errcode CPUCore::opRETD( u16 op1 ) {
	if( depth > 0)
	  --depth;
	return jump( a(14), false, true );
}
CPUCore::errcode CPUCore::opRET_FIQ( u16 op1 ) {
	SR = SPSR[subSPSR(SSR_FIQ)];
	return jump( (SPSR[subSPSR(SPCH_FIQ)] << 16) | SPSR[subSPSR(SPCL_FIQ)] );
}
CPUCore::errcode CPUCore::opRET_IRQ( u16 op1 ) {
	SR = SPSR[subSPSR(SSR_IRQ)];
	return jump( (SPSR[subSPSR(SPCH_IRQ)] << 16) | SPSR[subSPSR(SPCL_IRQ)] );
}
CPUCore::errcode CPUCore::opRET_SWI( u16 op1 ) {
	SR = SPSR[subSPSR(SSR_SWI)];
	return jump(a(14));
}
CPUCore::errcode CPUCore::opRL( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opRR( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opRRC( u16 op1 ) {
	int n = op1 & 0x000f;
	u16 val = r(n);
	r(n, (u16)((val >> 1) | ((flagT()?1:0) << 15)));
	flagT( val & 0x0001 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSBC1( u16 op1 ) {
	return subADSBC( (op1 & 0x0f00) >> 8, r(op1 & 0x000f), flagT(), true );
}
CPUCore::errcode CPUCore::opSBC2( u16 op1, u16 op2 ) {
	return subADSBC( op1 & 0x000f, op2, flagT(), true );
}
CPUCore::errcode CPUCore::opSETSR( u16 op1 ) {
	switch( op1 & 0x0007 ) {
		case 0:
			flagFE( true );
			break;
		case 1:
			flagIE( true );
			break;
		case 2:
			flagTE( true );
			break;
		case 3:
			flagV( true );
			break;
		case 4:
			flagZ0( true );
			break;
		case 5:
			flagZ1( true );
			break;
		case 6:
			flagPM( true );
			break;
		default:
			assert( false );
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSLB( u16 op1 ) {
	int n = op1 & 0xf;
	u16 val = r(n);
	flagT( val & 0x100 );
	r(n, (u16)(val << 8));
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSR( u16 op1 ) {
	int n = op1 & 0x000f;
	flagT( r(n) & 0x0001 );
	r( n, (u16)(r(n) >> 1) );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSRA( u16 op1 ) {
	int n = op1 & 0xf;
	u16 val = r(n);
	r(n, (u16)((val >> 1) | (val & 0x8000)));
	flagT( val & 1 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSRB( u16 op1 ) {
	int n = op1 & 0xf;
	u16 val = r(n);
	flagT( val & 0x80 );
	r(n, (u16)(val >> 8));
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSUB1( u16 op1 ) {
	return subADSBC( (op1 & 0xf00) >> 8, r(op1 & 0xf), false, true );
}
CPUCore::errcode CPUCore::opSUB2( u16 op1 ) {
	return subADSBC( (op1 & 0xf00) >> 8, (u16)(op1 & 0x7f), false, true );
}
CPUCore::errcode CPUCore::opSUB3( u16 op1 ) {
	int n = ((op1 & 0x0700) >> 8) + 8;
	int i = op1 & 0x000f;
	u22 val = (u22)(a(n) - (u22)r(i));
	a( n, val );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSUB4( u16 op1, u16 op2 ) {
	int n = (op1 & 0x0007) + 8;
	u22 val = (u22)(a(n) - (u22)op2);
	a( n, val );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSUB5( u16 op1 ) {
	int n = ((op1 & 0x0700) >> 8) + 8;
	u22 val = (u22)(a(n) - (u22)(op1 & 0x001f));
	a( n, val );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opSWI( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opSYS( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opTST1( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int i = op1 & 0x000f;
	u16 val = r(n) & r(i);
	flagT( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opTST2( u16 op1 ) {
	u16 imm = op1 & (u16)0x00ff;
	u16 val = r(0) & imm;
	flagT( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opTST3( u16 op1, u16 op2 ) {
	int n = op1 & 0x000f;
	u16 val = r(n) & op2;
	flagT( val == 0 );
	return SUCCESS;
}
CPUCore::errcode CPUCore::opTSTSR( u16 op1 ) {
	switch( op1 & 0x0007 ) {
		case 0:
			flagT(!flagFE());
			break;
		case 1:
			flagT(!flagIE());
			break;
		case 2:
			flagT(!flagTE());
			break;
		case 3:
			flagT(!flagV());
			break;
		case 4:
			flagT(!flagZ0());
			break;
		case 5:
			flagT(!flagZ1());
			break;
		case 6:
			flagT(!flagPM());
			break;
		default:
			assert( false );
	}
	return SUCCESS;
}
CPUCore::errcode CPUCore::opXOR1( u16 op1 ) {
	int n = (op1 & 0x0f00) >> 8;
	int i = op1 & 0x000f;
	u16 val = r(n) ^ r(i);
	flagT( val == 0 );
	r(n, val);
	return SUCCESS;
}
CPUCore::errcode CPUCore::opXOR2( u16 op1 ) {
	u16 imm = op1 & (u16)0x00ff;
	u16 val = r(0) ^ imm;
	flagT( val == 0 );
	r(0, val);
	return SUCCESS;
}
CPUCore::errcode CPUCore::opXOR3( u16 op1, u16 op2 ) {
	int n = op1 & 0x000f;
	u16 val = r(n) ^ op2;
	flagT( val == 0 );
	r(n, val);
	return SUCCESS;
}
CPUCore::errcode CPUCore::opEABS( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEADD1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEADD1( u16 op1, u16 op2 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEADD2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEADD3( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEADD4( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEBK( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opECLD( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opECP1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opECP1( u16 op1, u16 op2 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opECP2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opECR( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEDECC( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEDIVQ( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEEXP( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEEXPC( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEFS16( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEFS8( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEFZ16( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEFZ8( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEINCC( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD1( u16 op1, u16 op2 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD3( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD4( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD5( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD5( u16 op1, u16 op2 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opELD6( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMAD1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMAD2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMAD3( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMAX1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMAX2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMIN1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMIN2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMLD1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMLD2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMLD3( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMSB1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMSB2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMSB3( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMUL1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMUL2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opEMUL3( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opENEG( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opENMSK( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opENOP( u16 op1 ) {return SUCCESS;}
CPUCore::errcode CPUCore::opER( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opERESR( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opERND( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opERPD( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opERPN( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opERPN( u16 op1, u16 op2 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opERPR( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opERPS( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opES( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESAT( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESD( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESEC( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESFT( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESFTA( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESFTD( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESFTL( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESLA1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESLA2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESLA( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESLC( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESRA1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESRA2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESRA8( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESRC( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESUB1( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESUB1( u16 op1, u16 op2 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESUB2( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESUB3( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opESUB4( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opETST( u16 op1 ) {return NOTIMPLEMENTED;}
CPUCore::errcode CPUCore::opReserved( u16 op1 ) {return RESERVED;}
CPUCore::errcode CPUCore::opInvalid( u16 op1 ) {return INVALID;}

