/*  Project: Open Source Gmini Tools
    file: MMU.cpp, abstraction of memory management unit

    Copyright (C) 2004  Michael R. Donat <michael@donat.org>
         with contributions by gromit' and Treyqae

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <cstdlib>
#include <fstream>
using namespace std;
#include <assert.h>
#include "types.h"
#include "MMIO.h"
#include "MMU.h"
#include "CommentMap.h"

MMU::MMU() {
   memory = new u8[MEMSIZE];
   memset( memory, 0x99, MEMSIZE );
   ram = new u8[RAMSIZE];
   memset( ram, 0xaa, RAMSIZE );
   memmap = new u8[MEMSIZE * 3];
   memset( memmap, 0, MEMSIZE * 3 );
   io = new MMIO( IOBASE );
}
MMU::~MMU() {
   delete memory;
   delete ram;
   delete io;
   delete memmap;
}
u16 MMU::cread16( u22 address ) {
   if( (address > IOBASE) && (address < (IOBASE + IOSIZE)) )
      return (*io).read16( address );
   return (u16)((memory[address] << 8) | memory[address+1]);
}
bool MMU::cwrite16( u22 address, u16 value ) {
   if( (address > IOBASE) && (address < (IOBASE + IOSIZE)) )
      return (*io).write16( address, value );
   memory[address + 0] = (u8)((value & 0xFF00) >>  8);
   memory[address + 1] = (u8)((value & 0x00FF) >>  0);
   return true;
}
u8 MMU::dread8( u22 address ) {
   addmap( address, READ8 );
   if( (address > IOBASE) && (address < (IOBASE + IOSIZE)) )
      return (*io).read8( address );
   return ram[address];
}
u16 MMU::dread16( u22 address ) {
   addmap( address, READ16 );
   if( (address > IOBASE) && (address < (IOBASE + IOSIZE)) )
      return (*io).read16( address );
   return (u16)((ram[address] << 8) | ram[address+1]);
}
bool MMU::dwrite8( u22 address, u8 value ) {
   addmap( address, WRITE8 );
   if( (address > IOBASE) && (address < (IOBASE + IOSIZE)) )
      return (*io).write8( address, value );
   ram[address] = value;
   return true;
}
bool MMU::dwrite16( u22 address, u16 value ) {
   addmap( address, WRITE16 );
   if( (address > IOBASE) && (address < (IOBASE + IOSIZE)) )
      return (*io).write16( address, value );
   ram[address + 0] = (u8)((value & 0xFF00) >>  8);
   ram[address + 1] = (u8)((value & 0x00FF) >>  0);
   return true;
}
u16 MMU::cread( u22 address, bool op32 ) {
   if( op32 )
      addmap( address, CODE32 );
   else
      addmap( address, CODE16 );
   return (u16)((memory[address] << 8) | memory[address+1]);
}
bool MMU::cwrite( u22 address, u8 value ) {
   addmap( address, FIRM );
   memory[address] = value;
   return true;
}
const u32 MMU::find( const u8* value, u8 length ) {
	for( u32 address = 0; MEMSIZE > (address + length); ++address )
		if( memory[address] == value[0] ) {
			u8 i = 1;
			for(; i < length && memory[address + i] == value[i]; ++i );
			if( i == length ) 
				return address;
		}
	return 0xFFFFFFFF;
}
void MMU::prncstr( u22 address, u8 length ) {
	char* buf = new char[ length + 1 ];
	strncpy( buf, (char*)(memory + address), length );
	buf[length] = '\0';
	cout << "PRNCSTR: " << buf << endl;
	delete buf;
}
bool MMU::load( const char* filename, u32 baseaddress ) {
   loadbase = baseaddress;
   ifstream data( filename, ios::in | ios::binary );
   if( !data ) {
      cout << "Could not open data file: " << filename << endl;
      return false;
   }
   u22 i = loadbase;
   u8 ch;
   while(!(data.eof())) {
      ch = data.get();
      cwrite( i++, ch );
   }
   loadsize = i - loadbase;
   data.close();
   cout << "Loaded " << loadsize << " bytes." << endl;
   return true;
}
void MMU::findStrings() {
	comments->findStrings( memory, loadsize, loadbase );
}
bool MMU::isData( u22 address ) {

/*   return (                                    address <             loadbase ) ||
          (address >= (       0 + loadbase) && address < (  0x0100 + loadbase)) ||
          (address >= (  0x0414 + loadbase) && address < (  0x0490 + loadbase)) ||
          (address >= ( 0x0f9f4 + loadbase) && address < ( 0x12000 + loadbase)) ||
          (address >= ( 0x4cf2a + loadbase) && address < (loadbase + loadsize)) ||
          (address >= (loadbase + loadsize)                                   );*/
	return false;
}
bool MMU::dump( char* datafile ) {
   mapData();
   ofstream data( datafile, ios::out | ios::binary );
   if( !data ) {
      cout << "Could not open map file: " << datafile << endl;
      return false;
   }
   data.write( (char*)memmap, MEMSIZE * 3 );
   data.close();
   return true;
}
bool MMU::dumpmem( char* datafile ) {
   ofstream data( datafile, ios::out | ios::binary );
   if( !data ) {
      cout << "Could not open memory file: " << datafile << endl;
      return false;
   }
   data.write( (char*)memory, MEMSIZE );
   data.close();
   return true;
}
bool MMU::dumpram( char* datafile ) {
   ofstream data( datafile, ios::out | ios::binary );
   if( !data ) {
      cout << "Could not open code file: " << datafile << endl;
      return false;
   }
   data.write( (char*)ram, RAMPAGE );
   data.close();
   return true;
}
void MMU::mapData() {
   for( u22 i = 0; i != MEMSIZE; ++i )
      if(isData(i)) addmap(i, DATA);
}
void MMU::addmap( u22 address, access type ) {
   int r = 2;
   int w = 1;
   int c = 0;
/*
Red      Green    Blue
00000000 00000000 00000000
jxfd12fd jwfd12fd jrfd12fd
r  xx    a   ww   a   rr
*/
   switch( type ) {
   //READ16, READ8, WRITE16, WRITE8,
   //CODE16, CODE8, FIRM,    JUMP,
   //DATA };
      case READ16:
         memmap[( address   *3)+r] |= 0x48;
         memmap[((address+1)*3)+r] |= 0x48;
         break;
      case READ8:
         memmap[( address   *3)+r] |= 0x44;
         break;
      case WRITE16:
         memmap[( address   *3)+w] |= 0x48;
         memmap[((address+1)*3)+w] |= 0x48;
         break;
      case WRITE8:
         memmap[( address   *3)+w] |= 0x44;
         break;
      case CODE32:
         memmap[((address-2)*3)+c] |= 0x4c;
         memmap[((address-1)*3)+c] |= 0x4c;
         memmap[( address   *3)+c] |= 0x4c;
         memmap[((address+1)*3)+c] |= 0x4c;
         break;
      case CODE16:
         memmap[( address   *3)+c] |= 0x44;
         memmap[((address+1)*3)+c] |= 0x44;
         break;
      case RJUMP:
         memmap[(address*3)+c] |= 0x80;
         memmap[(address*3)+w] |= 0x00;
         memmap[(address*3)+r] |= 0x00;
         break;
      case AJUMP:
         memmap[(address*3)+c] |= 0x00;
         memmap[(address*3)+w] |= 0x80;
         memmap[(address*3)+r] |= 0x80;
         break;
      case FIRM:
         memmap[(address*3)+c] |= 0x22;
         memmap[(address*3)+w] |= 0x22;
         memmap[(address*3)+r] |= 0x22;
         break;
      case DATA:
         memmap[(address*3)+c] |= 0x11;
         memmap[(address*3)+w] |= 0x11;
         memmap[(address*3)+r] |= 0x11;
         break;
   }
}

