/*  Project: Open Source Gmini Tools
    file: SymMap.cpp, mantains a dynamic symbol mapping

    Copyright (C) 2004  Michael R. Donat <michael@donat.org>
         with contributions by gromit' and Treyqae

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <fstream>
#include <stdio.h>
#include "SymMap.h"

SymMap* symbols;

SymMap::SymMap() {
   load( "./symbols.std" );
}

SymMap::~SymMap() {
   map<u32,Reference*>::const_iterator itr = symmap.begin();
   while( itr != symmap.end() ) {
      delete itr->second;
      ++itr;
   }
}

int SymMap::load( const char* mapfile ) {
	char buffer[BUFLEN];
	if (!mapfile)
		return 0;
	ifstream data( mapfile, ios::in );
	if( !data ) {
		cout << "Could not open map file: " << mapfile << endl;
		return -1;
	}
	cout << "Loading symbols from " << mapfile << ": ";
#ifdef __GMEMU_DEBUG__
	cout << endl;
#endif
	char symname[BUFLEN];
	char refsym[BUFLEN];
	char rtypename[BUFLEN];
	u32 addr = 0;
	data.getline( buffer, BUFLEN );
	int counter = 0;
	while( !data.eof() ) {
		if( strncmp( "0x", buffer, 2 ) == 0 ) {
			sscanf( buffer, "0x%X %s", &addr, symname );
			//         if( symname[0] != '|' )
			symmap[addr] = new Reference( symname );
			//         else { // "|module$$segment$$type|"
			//         }
#ifdef __GMEMU_DEBUG__
			cout << hex << addr << dec << " = \"" << symname << "\"" << endl;
#endif
			++counter;
		}
		else if( buffer[0] == ' ' ) { // References (line begins with single space) " blah 0xXXXXXX\n"
			u32 from;
			sscanf( buffer, "%s 0x%X", rtypename, &from );
			int index = 0;
			while( index < 4 && strcmp( Reference::typenames[index], rtypename ) != 0 )
				++index;
			if( index < 4 )
				addref( from, addr, (Reference::access)index );
		}
		else if( strchr(buffer, '+') != NULL ) {
		// Symbol defined by offset of another
		//  referenced symbol must be defined earlier in the file
		//  only positive offsets allowed
		// "symbol+0xXXXX Symbol"
			char* plus = strchr( buffer, '+' );
			strncpy( refsym, buffer, plus - buffer );
			refsym[ plus - buffer ] = '\0';
			sscanf( plus + 1, "0x%x", &addr );
			strcpy( symname, strchr( buffer, ' ' ) + 1 );
			addr += addrof( refsym );
			if( addrof(refsym) != 0xFFFFFFF0 ) {
				symmap[addr] = new Reference( symname );
#ifdef __GMEMU_DEBUG__
				cout << hex << addr << dec << " = \"" << symname << "\"" << endl;
#endif
				++counter;
			}
				
		}
		data.getline( buffer, BUFLEN );
	}
	cout << counter << " symbols loaded." << endl;
	data.close();
	return 0;
}

const char *SymMap::get( u32 addr ) {
   if( symmap.find( addr ) != symmap.end() )
      return symmap[addr]->getname();
   return NULL;
}

void SymMap::addref( u32 from, u32 to, Reference::access type ) {
	if( symmap.find( to ) == symmap.end() ) {
		char name[24];
		switch( type ) {
			case Reference::DATA:
				sprintf( name, "data_%06X", to );
				break;
			case Reference::RJUMP:
				sprintf( name, "rel_%06X", to );
				break;
			case Reference::AJUMP:
				sprintf( name, "abs_%06X", to );
				break;
			case Reference::CALL:
				sprintf( name, "sub_%06X", to );
				break;
		}
		symmap[to] = new Reference( name );
	}
	if( symmap.find( to ) != symmap.end() )
		symmap[to]->add( from, type );
}

const char* SymMap::getref( u32 addr, char* buffer ) {
   if( symmap.find( addr ) != symmap.end() ) {
      symmap[addr]->getrefs( buffer );
      return buffer;
   }
   return NULL;
}

const u32 SymMap::addrof( char* symbol ) {
   map<u32,Reference*>::const_iterator itr = symmap.begin();
   while( itr != symmap.end() && (strcmp( symbol, itr->second->getname() ) != 0)) {
      ++itr;
   }
   if( itr != symmap.end() )
      return itr->first;
   return 0xFFFFFFF0;
}

const int SymMap::dump( const char* file ) {
   ofstream data( file, ios::out );
   if( !data ) {
      cout << "Could not open symbol file: " << file << endl;
      return -1;
   }
   char* buffer = new char[ 128 * 1024 ];
   map<u32,Reference*>::const_iterator itr = symmap.begin();
   while( itr != symmap.end() ) {
      int length = sprintf( buffer, "0x%06X %s\n", itr->first, itr->second->getname() );
      length += itr->second->getrefs( buffer + length );
      data.write( buffer, length );
      ++itr;
   }
   delete buffer;
   data.close();
   return 0;
}

