/*  Project: Open Source Gmini Tools
    file: decodeOP.cpp, functions to parse binary instruction data.

    Copyright (C) 2004  Michael R. Donat <michael@donat.org>
         with contributions by gromit' and Treyqae

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "types.h"
#include "decodeOP.h"

s32 sext4to32( u32 val ) {
   if( val & 0x00000008 )
      val |= 0xfffffff0;
   return (s32)val;
}
s32 sext5to32( u32 val ) {
   if( val & 0x00000010 )
      val |= 0xffffffe0;
   return (s32)val;
}
s16 sext8to16( u32 val ) {
   if( val & 0x00000080 )
      val |= 0xffffff00;
   return (s16)val;
}
s32 sext8to32( u32 val ) {
   if( val & 0x00000080 )
      val |= 0xffffff00;
   return (s32)val;
}
s32 sext11to32( u32 val ) {
   if( val & 0x00000400 )
      val |= 0xfffff800;
   return (s32)val;
}
s32 sext13to32( u32 val ) {
   if( val & 0x00001000 )
      val |= 0xffffe000;
   return (s32)val;
}
s32 sext16to32( u32 val ) {
   if( val & 0x00008000 )
      val |= 0xffff0000;
   return (s32)val;
}

/*
//_0__1__2__3__4__5__6__7__8__9__a__b__c__d__e__f__////////////////
// 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, //00 (  0 -  15)
// 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, //10 ( 16 -  31)
// 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, //20 ( 32 -  47)
// 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, //30 ( 48 -  63)
// 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, //40 ( 64 -  7f)
// 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, //50 ( 80 -  95)
// 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, //60 ( 96 - 111)
// 2, 2, 2, 2, 2, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 4, //70 (112 - 127)
// 4, 4, 4, 4, 4, 2, 2, 2, 4, 4, 2, 2, 2, 2, 2, 2, //80 (128 - 143)
// 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, //90 (144 - 159)
// 2, 2, 2, 2, 2, 4, 4, 4, 4, 4, 4, 2, 2, 2, ...   //a0 (160 - 281)
//_0__1__2__3__4__5__6__7__8__9__a__b__c__d__e__f__////////////////
*/
int opsize( int mnnum ) {
   if( mnnum < 0x31 )
      return 2;
   else if( mnnum < 0x40 )
      return 4;
   else if( mnnum < 0x75 )
      return 2;
   else if( mnnum < 0x79 )
      return 4;
   else if( mnnum < 0x7f )
      return 2;
   else if( mnnum < 0x85 )
      return 4;
   else if( mnnum < 0x88 )
      return 2;
   else if( mnnum < 0x8a )
      return 4;
   else if( mnnum < 0xa5 )
      return 2;
   else if( mnnum < 0xab )
      return 4;
   return 2;
}

int decodeOP( u16 op ) {
   if( !(op &0x8000) ) {
      if( !(op &0x4000) ) {
         if( !(op &0x2000) ) {
            if( !(op &0x1000) )
               return (op & 0x0080) >> 7 ;
            return 2;
         }
         return 3 + ((op & 0x1000) >> 12);
      }
      if( !(op &0x2000) )
         return 5 + (((op & 0x1000) >> 11) | ((op & 0x0080) >> 7));
      return 9 + (((op & 0x1800) >> 10) | ((op & 0x0080) >> 7));
   }
   if( !(op &0x4000) ) {
      if( !(op &0x2000) ) {
         if( !(op &0x1000) ) {
            if( (op & 0x00e0) != 0x00e0 )
               return 17 + ((op & 0x00f0) >> 4);
            if( !(op & 0x0010) ) {
               if( !(op & 0x0800) ) {
                  if( (op & 0x0600) != 0x0600 )
                     return 31 + ((op & 0x0700) >> 8);
                  return 37 + (((op & 0x0100) >> 7) | ((op & 0x0008) >> 3));
               }
               return 41 + ((op & 0x0700) >> 8);
            }
            if( !(op & 0x0e00) ) {
               if( !(op & 0x0100) )
                  return 49;
               return 50 + ((op & 0x0008) >> 3);
            }
            if( (op & 0x0e00) != 0x0e00 )
               return 52 + ((op & 0x0f00) >> 8) - 2;
            return 64;
         }
         if( !(op & 0x0800) )
            return 65;
         if( (op & 0x0600) != 0x0600 )
            return 66 + ((op & 0x0700) >> 8);
         if( !(op & 0x0100) ) {
            if( !(op & 0x0080) ) {
               if( !(op & 0x0060) )
                  return 72 + ((op & 0x0018) >> 3);
               return 76 + ((op & 0x0040) >> 6);
            }
            if( !(op & 0x0040) ) {
               if( !(op & 0x0020) ) {
                  if( (op & 0x0018) != 0x00180 )
                     return 78 + ((op & 0x0018) >> 3);
                  return 81 + (op & 0x0007);
               }
               if( !(op & 0x0010) ) {
                  if( !(op & 0x0008) )
                     return 89 + (op & 0x0007);
                  if( !(op & 0x0004) ) {
                     if( !(op & 0x0002) )
                        return 97;
                     return 98 + (op & 0x0001);
                  }
                  return 100;
               }
               if( !(op & 0x0008) )
                  return 101 + (op & 0x0007);
               if( !(op & 0x0004) ) {
                  if( !(op & 0x0002) )
                     return 109;
                  return 110 + (op & 0x0001);
               }
               return 112;
            }
            if( !(op & 0x0020) )
               return 113;
            if( !(op & 0x0010) )
               return 114;
            return 115 + ((op & 0x0008) >> 3);
         }
         return 117 + ((op & 0x00c0) >> 6);
      }
      if( !(op & 0x1000) ) {
         if( !(op & 0x0080) ) {
            if( !(op & 0x0040) ) {
               if( !(op & 0x0020) ) {
                  if( !(op & 0x0010) )
                     return 121 + ((op & 0x0008) >> 3);
                  return 123 + ( ((op & 0x0800) >> 10) | ((op & 0x0008) >> 3) );
               }
               if( !(op & 0x0010) )
                  return 127 + ((op & 0x0008) >> 3);
               return 129 + ( ((op & 0x0800) >> 10) | ((op & 0x0008) >> 3) );
            }
            if( !(op & 0x0800) )
               return 133;
            return 134 + ((op & 0x0020) >> 5);
         }
         if( !(op & 0x0040) )
            return 136 + ((op & 0x0800) >> 11);
         if( !(op & 0x0020) ) {
            if( !(op & 0x0010) )
               return 138 + ((op & 0x0800) >> 11);
            return 140 + ( ((op & 0x0800) >> 10) | ((op & 0x0008) >> 3) );
         }
         if( !(op & 0x0010) ) {
            if( !(op & 0x0008) )
               return 144;
            return 145 + ((op & 0x0800) >> 11);
         }
         if( !(op & 0x0008) )
            return 147;
         return 148 + ((op & 0x0800) >> 11);
      }
      return 150;
   }
   if( !(op & 0x2000) ) {
      if( !(op & 0x1800) ) {
         if( (op & 0x0600) != 0x0600 )
            return 151 + ((op & 0x0600) >> 9);
         return 154 + ((op & 0x0080) >> 7);
      }
      return 156 + ((op & 0x1c00) >> 10) - 2;
   }
   if( !(op & 0x1e00) )
      return 162 + ((op & 0x0010) >> 4);
   return 164 + decodeCOP( op & (u16)0x1fff );
}

int decodeCOP( u16 op ) {
   if( !(op & 0x1000) ) {
      if( !(op & 0x0800) ) {
         if( !(op & 0x0400) ) {
            if( !(op & 0x0200) )
               return 0;
            if( !(op & 0x0100) )
               return 1;
            if( !(op & 0x0080) )
               return 2;
            return 3 + ((op & 0x0060) >> 5 );
         }
         if( !(op & 0x0200) )
            return 8 + ((op & 0x0001) >> 4 );
         if( !(op & 0x0010) )
            return 9;
         if( !(op & 0x0008) )
            return 10;
         return 11 + ((op & 0x0100) >> 8 );
      }
      if( !(op & 0x0400) ) {
         if( !(op & 0x0200) ) {
            if( !(op & 0x0100) )
               return 13 + ((op & 0x0018) >> 3 );
            if( (op & 0x0018) != 0x0018 )
               return 17 + ((op & 0x0018) >> 3 );
            return 20 + ((op & 0x0060) >> 5 );
         }
         if( !(op & 0x0100) )
            return 24 + ((op & 0x0010) >> 4 );
         if( !(op & 0x0010) )
            return 26;
         if( !(op & 0x0040) )
            return 27 + ((op & 0x0020) >> 5 );
         if( !(op & 0x0020) )
            return 29;
         if( !(op & 0x0008) )
            return 30;
         return 31;
      }
      if( !(op & 0x0200) ) {
         if( !(op & 0x0100) )
            return 32 + ((op & 0x0010) >> 4 );
         if( !(op & 0x0010) )
            return 34;
         if( !(op & 0x0040) )
            return 35 + ((op & 0x0020) >> 5 );
         if( !(op & 0x0020) )
            return 37;
         return 38 + ((op & 0x0008) >> 3 );
      }
      if( !(op & 0x0100) ) {
         if( !(op & 0x0080) )
            return 40 + ((op & 0x0040) >> 6 );
         if( !(op & 0x0040) )
            return 42;
         return 43 + ((op & 0x0010) >> 4 );
      }
      if( !(op & 0x0010) ) {
         if( !(op & 0x0008) )
            return 45;
         return 46 + ((op & 0x0080) >> 7 );
      }
      if( !(op & 0x0008) )
         return 48;
      return 49 + ((op & 0x0080) >> 7 );
   }
   if( !(op & 0x0800) ) {
      if( !(op & 0x0400) ) {
         if( !(op & 0x0200) )
            return 51 + ((op & 0x0010) >> 4 );
         if( !(op & 0x0010) )
            return 53;
         if( !(op & 0x0008) )
            return 54;
         if( !(op & 0x0100) )
            return 55 + ((op & 0x0080) >> 7 );
         return 57;
      }
      if( !(op & 0x0200) ) {
         if( !(op & 0x0100) ) {
            if( !(op & 0x0080) ) {
               if( !(op & 0x0040) )
                  return 58 + ((op & 0x0030) >> 4 );
               if( !(op & 0x0020) )
                  return 62;
               return 63 + ((op & 0x0010) >> 4 );
            }
            return 65 + ((op & 0x0040) >> 6 );
         }
         return 67 + ((op & 0x00c0) >> 6 );
      }
      return 71 + ((op & 0x0100) >> 8 );
   }
   if( !(op & 0x0400) ) {
      if( !(op & 0x0200) )
         return 73;
      if( !(op & 0x0100) ) {
         if( !(op & 0x0080) )
            return 74 + ((op & 0x0010) >> 4 );
         return 76;
      }
      if( !(op & 0x0080) ) {
         if( !(op & 0x0040) )
            return 77 + ((op & 0x0020) >> 5 );
         if( !(op & 0x0020) )
            return 79;
         if( !(op & 0x0010) )
            return 80;
         if( !(op & 0x0002) )
            return 81 + (op & 0x0001);
         if( !(op & 0x0004) )
            return 83 + (op & 0x0001);
         return 85;
      }
      return 86 + ((op & 0x0040) >> 6 );
   }
   if( !(op & 0x0200) ) {
      if( !(op & 0x0100) ) {
         if( !(op & 0x0080) ) {
            if( !(op & 0x0040) )
               return 88 + ((op & 0x0020) >> 5 );
            if( !(op & 0x0020) )
               return 90;
            if( !(op & 0x0010) )
               return 91;
            return 92 + ((op & 0x0008) >> 3 );
         }
         if( !(op & 0x0040) )
            return 94 + ((op & 0x0020) >> 5 );
         if( !(op & 0x0020) )
            return 96;
         if( !(op & 0x0010) )
            return 97;
         if( !(op & 0x0008) )
            return 98;
         if( !(op & 0x0004) )
            return 99;
         return 100;
      }
      return 101 + ((op & 0x0080) >> 7 );
   }
   if( !(op & 0x0100) )
      return 103 + ((op & 0x0080) >> 7 );
   if( !(op & 0x0080) ) {
      if( !(op & 0x0040) ) {
         if( !(op & 0x0020) ) {
            if( !(op & 0x0010) )
               return 105;
            return 106 + ((op & 0x0008) >> 3 );
         }
         return 108;
      }
      return 109 + ((op & 0x0020) >> 5 );
   }
   return 111 + ((op & 0x0070) >> 4 );
}

