#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#define _GNU_SOURCE
#include <fnmatch.h>
#undef _GNU_SOURCE

#define NB_FIRST_BYTES 32
#define NB_CRC_BYTES 3

typedef struct
{
	char *symbol;
	unsigned int crc_nb_bytes;
	unsigned int crc;
	/* ... */
} signature_t;

typedef struct treenode_s
{
	char code;
	struct treenode_s *childs;
	signature_t *signature;
	struct treenode_s *next;
} treenode_t;

typedef struct
{
	unsigned int size;
	treenode_t *root;
} tree_t;

int crc16(char *ptr, int count)
{
	int crc, i;

	crc = 0;
	while (--count >= 0)
	{
		crc = crc ^ (int)*ptr++ << 8;
		for (i = 0; i < 8; ++i)
		{
			if (crc & 0x8000)
			{
				crc = crc << 1 ^ 0x1021;
			}
			else
			{
				crc = crc << 1;
			}
		}
	}
	return (crc & 0xFFFF);
}

tree_t *tree_create(void)
{
	tree_t *tree;
	
	if (!(tree = (tree_t*)malloc(sizeof(tree_t))))
	{
		return NULL;
	}
	if (!(tree->root = (treenode_t*)malloc(sizeof(treenode_t))))
	{
		return NULL;
	}
	return tree;
}


void treenode_add(treenode_t *tn, char *symbol, char *buffer, int len, int position)
{
	char c;
	treenode_t *child;

	if ((position < NB_FIRST_BYTES-1) && (position < len))
	{
		c = buffer[position];
		/* search through the child list to find which has the right code */
		for (child = tn->childs; (child && child->code != c); child = child->next)
			;
		if (!child)
		{
			/* new child inserted at the beginning of the list */
			if (!(child = (treenode_t*)malloc(sizeof(treenode_t))))
			{
				printf("No more memory !\n");
				exit(EXIT_FAILURE);
			}
			child->signature = NULL;
			child->childs = NULL;
			child->code = c;
			child->next = tn->childs;
			tn->childs = child;
		}
		treenode_add(child, symbol, buffer, len, position+1);
		return;
	}
		
	if (tn->signature)
	{
		printf("Conflicting symbols... need to refine the algorithm !\n");
		exit(EXIT_FAILURE);
	}
	
	if (!(tn->signature = (signature_t*)malloc(sizeof(signature_t))))
	{
		printf("No more memory !\n");
		exit(EXIT_FAILURE);
	}
	tn->signature->symbol = strdup(symbol);
	tn->signature->crc = 0;
	tn->signature->crc_nb_bytes = 0;
}

void tree_add(tree_t *tree, char *symbol, char *buffer, int len)
{
	treenode_add(tree->root, symbol, buffer, len, 0);
}

void tree_fill(tree_t *tree, char *directory)
{
	struct dirent **namelist;
	int nb_entries;
	int i;

        nb_entries = scandir(directory, &namelist, 0, alphasort);

        if (nb_entries < 0)
        {
		printf("Cannot scan directory !\n");
		exit(EXIT_FAILURE);
	}
        else
        {
                for (i = 2; i < nb_entries; i++)
		{
			if (fnmatch("*.o", namelist[i]->d_name, FNM_CASEFOLD) == 0)
			{
				FILE *fd;
				char buffer[NB_FIRST_BYTES];

				printf("Adding symbol [%s]\n", namelist[i]->d_name);
				fd = fopen(namelist[i]->d_name, "r");
				fread(buffer, 1, NB_FIRST_BYTES, fd);
				fclose(fd);

				tree_add(tree, namelist[i]->d_name, buffer, NB_FIRST_BYTES);
				tree->size++;
			}
		}
	}
}

char *tree_match(tree_t *tree, char *buffer, int len)
{
	treenode_t *tn = tree->root;
	treenode_t *child;
	int position = 0;
	char c;
	
	while (position < len)
	{
		c = buffer[position];
		for (child = tn->childs; (child && child->code != c); child = child->next)
			;
		if (!child)
		{
			return NULL;
		}
		tn = child;
		position++;
	}
	if (child->signature)
	{
		return child->signature->symbol;
	}
	else
	{
		return NULL;
	}
}



void tree_print_rec(treenode_t *tn, char* buffer, int len, int position)
{	
	treenode_t *child;
	int i;

	if (position != -1)
	{
		buffer[position] = (tn->code & 0xff);
	}

	if (tn->signature)
	{
		buffer[position+1] = 0;
		/* TODO improve with crc etc. */
		for (i = 0; i < position; i++)
		{
			printf("%02x", (buffer[i] & 0xff));
		}
		printf(" %s %04x\n", tn->signature->symbol, crc16(tn->signature->symbol, strlen(tn->signature->symbol)));
	}
	
	for (child = tn->childs; (child && child->code != tn->code); child = child->next)
	{
		tree_print_rec(child, buffer, len, position+1);
	}
}

void tree_print(tree_t *tree)
{
	char buffer[NB_FIRST_BYTES+1];
	tree_print_rec(tree->root, buffer, NB_FIRST_BYTES, -1);
}

void tree_write_rec(treenode_t *tn, FILE *fd, char* buffer, int len, int position)
{	
	treenode_t *child;
	int i;

	if (position != -1)
	{
		buffer[position] = (tn->code & 0xff);
	}

	if (tn->signature)
	{
		memset(buffer + position + 1, 0, len - position - 1);
		/* TODO improve with crc etc. */
		for (i = 0; i < position; i++)
		{
			fprintf(fd, "%02x", (buffer[i] & 0xff));
		}
		fprintf(fd, " %s\n", tn->signature->symbol);
	}
	
	for (child = tn->childs; (child && child->code != tn->code); child = child->next)
	{
		tree_write_rec(child, fd, buffer, len, position+1);
	}
}

void tree_write(tree_t *tree, char *filename)
{
	FILE *fd;
	char buffer[NB_FIRST_BYTES+1];
	
	fd = fopen(filename, "w");
	tree_write_rec(tree->root, fd, buffer, NB_FIRST_BYTES, -1);
	fclose(fd);
}

int main(int argc, char **argv)
{
	tree_t *tree;
	//char buffer[NB_FIRST_BYTES];
	//FILE *fd;
	
	if (!(tree = tree_create()))
	{
		printf("No more memory !\n");
		exit(EXIT_FAILURE);
	}
	
	tree->size = 0;
	tree_fill(tree, argv[1]);

	//fd = fopen("sprintf.o", "r");
	//fread(buffer, 1, NB_FIRST_BYTES, fd);
	//fclose(fd);
	//printf("%s\n", tree_match(tree, buffer, NB_FIRST_BYTES)); 
	//buffer[2] = 0;
	//printf("%s\n", tree_match(tree, buffer, NB_FIRST_BYTES)); 
	//printf("\n");
	tree_print(tree);
	tree_write(tree, "tree.sig");

	return EXIT_SUCCESS;
}

