/*  Project: Open Source Gmini Tools
    file: formatCOP.cpp, functions to format instruction mnemonics
          for the MAC2424 DSP of the TCC730

    Copyright (C) 2004  Michael R. Donat <michael@donat.org>
         with contributions by gromit' and Treyqae

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include "types.h"
#include "macargs.h"
#include "decodeOP.h"
#include "formatCOP.h"

char COPformat[256];

char *formatCOP( int mnnum, u16 op1, u16 op2 ) {
   /* operands */
   s32 imm4, imm5, imm16;
   u8 adr2, adr5,
      t1,t2,t3,t4, t5, t6;
   u16 adr16;
   switch( mnnum ) {
      case 0: /* ECLD */
         sprintf( COPformat, "ECLD" );
         break;
      case 1: /* ELD mg,#imm:16 */
         t1 = (u8)((op1 & 0x00f8) >> 3 );
         imm16 = sext16to32( op2 );
         sprintf( COPformat, "ELD %s, #%d", mg[t1], imm16 );
         break;
      case 2: /* EMOD0 An,#imm:16 */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         t2 = (u8)((op1 & 0x0018) >> 3 );
         imm16 = sext16to32( op2 );
         sprintf( COPformat, "%s %s, #%d", emod0[t1], An[t2], imm16 );
         break;
      case 3: /* ELD mgx,#imm:16 */
         t1 = (u8)((op1 & 0x0018) >> 3 );
         imm16 = sext16to32( op2 );
         sprintf( COPformat, "ELD %s, #%d", mgx[t1], imm16 );
         break;
      case 4: /* ERPN rpi, #imm:16 */
         t1 = (u8)((op1 & 0x0018) >> 3 );
         imm16 = sext16to32( op2 );
         sprintf( COPformat, "ERPN %s, #%d", rpi[t1], imm16 );
         break;
      case 5: /* ELD An,adr:16 */
         t1 = (u8)((op1 & 0x0018) >> 3 );
         adr16 = (u16)op2;
         sprintf( COPformat, "ELD %s, #%d", An[t1], adr16 );
         break;
      case 6: /* ELD adr:16,An */
         t1 = (u8)((op1 & 0x0018) >> 3 );
         adr16 = (u16)op2;
         sprintf( COPformat, "ELD #%d, %s", adr16, An[t1] );
         break;
      case 7: /* EMAD Mi, XiYi, mgx,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0180) >> 7 );
         t3 = (u8)((op1 & 0x0060) >> 5 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EMAD %s, %s, %s, @%s", Mi[t1], XiYi[t2], mgx[t3], rps[t4] );
         break;
      case 8: /* EMSB Mi, XiYi, mgx,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0180) >> 7 );
         t3 = (u8)((op1 & 0x0060) >> 5 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EMSB %s, %s, %s, @%s", Mi[t1], XiYi[t2], mgx[t3], rps[t4] );
         break;
      case 9: /* EMLD Mi, XiYi, mgx,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0180) >> 7 );
         t3 = (u8)((op1 & 0x0060) >> 5 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EMLD %s, %s, %s, @%s", Mi[t1], XiYi[t2], mgx[t3], rps[t4] );
         break;
      case 10: /* EMUL XiYi, mgx,@rps */
         t1 = (u8)((op1 & 0x0180) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EMUL %s, %s, @%s", XiYi[t1], mgx[t2], rps[t3]);
         break;
      case 11: /* EADD Mi,P, mgx,@rps */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EADD %s, P, %s, @%s", Mi[t1], mgx[t2], rps[t3] );
         break;
      case 12: /* ESUB Mi,P, mgx,@rps */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ESUB %s, P, %s, @%s", Mi[t1], mgx[t2], rps[t3] );
         break;
      case 13: /* EADD Mi,P, An,@rps */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EADD %s, P, %s, @%s", Mi[t1], An[t2], rps[t3] );
         break;
      case 14: /* ESUB Mi,P, An,@rps */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ESUB %s, P, %s, @%s", Mi[t1], An[t2], rps[t3] );
         break;
      case 15: /* ELD Mi,P, An,@rps */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, P, %s, @%s", Mi[t1], An[t2], rps[t3] );
         break;
      case 16: /* ELD Mi,P, mgx,@rps */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, P, %s, @%s", Mi[t1], mgx[t2], rps[t3] );
         break;
      case 17: /* EADD Mi,P, @rpd,mga */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)(op1 & 0x0007);
         t3 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "EADD %s, P, @%s, %s", Mi[t1], rpd[t2], mga[t3] );
         break;
      case 18: /* ESUB Mi,P, @rpd,mga */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)(op1 & 0x0007);
         t3 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ESUB %s, P, @%s, %s", Mi[t1], rpd[t2], mga[t3] );
         break;
      case 19: /* ELD Mi,P, @rpd,mga */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)(op1 & 0x0007);
         t3 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ELD %s, P, @%s, %s", Mi[t1], rpd[t2], mga[t3] );
         break;
      case 20: /* EADD Mi,P, @rpd,P */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EADD %s, P, @%s, P", Mi[t1], rpd[t2] );
         break;
      case 21: /* ESUB Mi,P, @rpd,P */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ESUB %s, P, @%s, P", Mi[t1], rpd[t2] );
         break;
      case 22: /* ELD Mi,P, @rpd,P */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, P, @%s, P", Mi[t1], rpd[t2] );
         break;
      case 23: /* reserved */
         sprintf( COPformat, "reserved" );
         break;
      case 24: /* EADD Ai,Mi, mgx,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)((op1 & 0x0060) >> 5 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EADD %s, %s, %s, @%s", Ai[t1], Mi[t2], mgx[t3], rps[t4] );
         break;
      case 25: /* ESUB Ai,Mi, mgx,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)((op1 & 0x0060) >> 5 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ESUB %s, %s, %s, @%s", Ai[t1], Mi[t2], mgx[t3], rps[t4] );
         break;
      case 26: /* ELD Ai,Mi, mgx,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)((op1 & 0x0060) >> 5 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, %s, %s, @%s", Ai[t1], Mi[t2], mgx[t3], rps[t4] );
         break;
      case 27: /* EADD Ai,Mi, Mi,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EADD %s, %s, %s, @%s", Ai[t1], Mi[t2], Mi[t2], rps[t3] );
         break;
      case 28: /* ESUB Ai,Mi, Mi,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ESUB %s, %s, %s, @%s", Ai[t1], Mi[t2], Mi[t2], rps[t3] );
         break;
      case 29: /* ELD Ai,Mi, Mi,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, %s, %s, @%s", Ai[t1], Mi[t2], Mi[t2], rps[t3] );
         break;
      case 30: /* ELD Pi,@rps */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, @%s", Pi[t1], rps[t2] );
         break;
      case 31: /* reserved */
         sprintf( COPformat, "reserved" );
         break;
      case 32: /* EADD Ai,Mi, @rpd,mga */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)(op1 & 0x0007);
         t4 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "EADD %s, %s, @%s, %s", Ai[t1], Mi[t2], rpd[t3], mga[t4] );
         break;
      case 33: /* ESUB Ai,Mi, @rpd,mga */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)(op1 & 0x0007);
         t4 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ESUB %s, %s, @%s, %s", Ai[t1], Mi[t2], rpd[t3], mga[t4] );
         break;
      case 34: /* ELD Ai,Mi, @rpd,mga */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)(op1 & 0x0007);
         t4 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ELD %s, %s, @%s, %s", Ai[t1], Mi[t2], rpd[t3], mga[t4] );
         break;
      case 35: /* EADD Ai,Mi, @rpd,P */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EADD %s, %s, @%s, P", Ai[t1], Mi[t2], rpd[t3] );
         break;
      case 36: /* ESUB Ai,Mi, @rpd,P */
         t1 = (u8)((op1 & 0x0008) >> 3);
         t2 = (u8)((op1 & 0x0080) >> 7);
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ESUB %s, %s, @%s, P", Ai[t1], Mi[t2], rpd[t3] );
         break;
      case 37: /* ELD Ai,Mi, @rpd,P */
         t1 = (u8)((op1 & 0x0008) >> 3);
         t2 = (u8)((op1 & 0x0080) >> 7);
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, %s, @%s, P", Ai[t1], Mi[t2], rpd[t3] );
         break;
      case 38: /* ELD @rpd,P */
         t1 = (u8)((op1 & 0x0080) >> 7);
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD @%s, %s", rpd[t1], Pi[t2] );
         break;
      case 39: /* reserved */
         sprintf( COPformat, "reserved" );
         break;
      case 40: /* EADD Ai,Ci, Cj,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0020) >> 5 );
         t3 = (u8)((op1 & 0x0010) >> 4 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EADD %s, %s, %s, @%s", Ai[t1], Ci[t2], Ci[t3], rps[t4] );
         break;
      case 41: /* ESUB Ai,Ci, Cj,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0020) >> 5 );
         t3 = (u8)((op1 & 0x0010) >> 4 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ESUB %s, %s, %s, @%s", Ai[t1], Ci[t2], Ci[t3], rps[t4] );
         break;
      case 42: /* ECP Ai,Ci, Cj,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0020) >> 5 );
         t3 = (u8)((op1 & 0x0010) >> 4 );
         t4 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ECP %s, %s, %s, @%s", Ai[t1], Ci[t2], Ci[t3], rps[t4] );
         break;
      case 43: /* EMAX Ai,Ci, Ci,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0020) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EMAX %s, %s, %s, @%s", Ai[t1], Ci[t2], Ci[t2], rps[t3] );
         break;
      case 44: /* EMIN Ai,Ci, Ci,@rps */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0020) >> 5 );
         t3 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "EMIN %s, %s, %s, @%s", Ai[t1], Ci[t2], Ci[t2], rps[t3] );
         break;
      case 45: /* ELD mg1,@rps */
         t1 = (u8)((op1 & 0x00e0) >> 5 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, @%s", mg1[t1], rps[t2] );
         break;
      case 46: /* ELD An,@rps */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, @%s", An[t1], rps[t2] );
         break;
      case 47: /* ELD srg,@rps */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, @%s", srg[t1], rps[t2] );
         break;
      case 48: /* ELD @rpd,mg1 */
         t1 = (u8)(op1 & 0x0007);
         t2 = (u8)((op1 & 0x00e0) >> 5 );
         sprintf( COPformat, "ELD @%s, %s", rpd[t1], mg1[t2] );
         break;
      case 49: /* ELD @rpd,An */
         t1 = (u8)(op1 & 0x0007);
         t2 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ELD @%s, %s", rpd[t1], An[t2] );
         break;
      case 50: /* ELD @rps,srg */
         t1 = (u8)(op1 & 0x0007);
         t2 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ELD @%s, %s", rps[t1], srg[t2] );
         break;
      case 51: /* EMAD Mi, XiYi, Xi,@rp01s, Yi,@rp3s */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0180) >> 7 );
         t3 = (u8)((op1 & 0x0040) >> 6 );
         t4 = (u8)(op1 & 0x0003);
         t5 = (u8)((op1 & 0x0020) >> 5 );
         t6 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMAD %s, %s, %s, @%s, %s, @%s", Mi[t1], XiYi[t2], Xi[t3], rp01s[t4], Yi[t5], rp3s[t6] );
         break;
      case 52: /* EMSB Mi, XiYi, Xi,@rp01s, Yi,@rp3s */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0180) >> 7 );
         t3 = (u8)((op1 & 0x0040) >> 6 );
         t4 = (u8)(op1 & 0x0003);
         t5 = (u8)((op1 & 0x0020) >> 5 );
         t6 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMSB %s, %s, %s, @%s, %s, @%s", Mi[t1], XiYi[t2], Xi[t3], rp01s[t4], Yi[t5], rp3s[t6] );
         break;
      case 53: /* EMLD Mi, XiYi, Xi,@rp01s, Yi,@rp3s */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0180) >> 7 );
         t3 = (u8)((op1 & 0x0040) >> 6 );
         t4 = (u8)(op1 & 0x0003);
         t5 = (u8)((op1 & 0x0020) >> 5 );
         t6 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMLD %s, %s, %s, @%s, %s, @%s", Mi[t1], XiYi[t2], Xi[t3], rp01s[t4], Yi[t5], rp3s[t6] );
         break;
      case 54: /* EMUL XiYi, Xi,@rp01s, Yi,@rp3s */
         t1 = (u8)((op1 & 0x0180) >> 7 );
         t2 = (u8)((op1 & 0x0040) >> 6 );
         t3 = (u8)(op1 & 0x0003);
         t4 = (u8)((op1 & 0x0020) >> 5 );
         t5 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMUL %s, %s, @%s, %s, @%s", XiYi[t1], Xi[t2], rp01s[t3], Yi[t4], rp3s[t5] );
         break;
      case 55: /* ELD Xi,@rp01s, Yi,@rp3s */
         t1 = (u8)((op1 & 0x0040) >> 6 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0020) >> 5 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "ELD %s, @%s, %s, @%s", Xi[t1], rp01s[t2], Yi[t3], rp3s[t4] );
         break;
      case 56: /* reserved */
         sprintf( COPformat, "reserved" );
         break;
      case 57: /* reserved */
         sprintf( COPformat, "reserved" );
         break;
      case 58: /* ESFT asr,asa */
         t1 = (u8)((op1 & 0x000c) >> 2 );
         t2 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "ESFT %s, %s", asr[t1], asa[t2] );
         break;
      case 59: /* ESFTA asr,asa */
         t1 = (u8)((op1 & 0x000c) >> 2 );
         t2 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "ESFTA %s, %s", asr[t1], asa[t2] );
         break;
      case 60: /* ESFTL asr,asa */
         t1 = (u8)((op1 & 0x000c) >> 2 );
         t2 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "ESFTL %s, %s", asr[t1], asa[t2] );
         break;
      case 61: /* ESFTD asr,asa */
         t1 = (u8)((op1 & 0x000c) >> 2 );
         t2 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "ESFTD %s, %s", asr[t1], asa[t2] );
         break;
      case 62: /* ELD SA,#imm:5 */
         imm5 = sext5to32( (u8)(op1 & 0x001f) );
         sprintf( COPformat, "ELD SA, #%d", imm5 );
         break;
      case 63: /* ENMSK SG,#imm:4 */
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ENMSK SG, #%d", imm4 );
         break;
      case 64: /* ELD srgd,srgs */
         t1 = (u8)((op1 & 0x000c) >> 2 );
         t2 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "ELD %s, %s", srg[t1], srg[t2] );
         break;
      case 65: /* ELD rpui,rpd1.adr:2 */
         t1 = (u8)(op1 & 0x000f);
         adr2 = (u8)((op1 & 0x0030) >> 4 );
         sprintf( COPformat, "ELD %s, RPD1.%d", rpui[t1], adr2 );
         break;
      case 66: /* ELD rpd1.adr:2,rpui */
         adr2 = (u8)((op1 & 0x0030) >> 4 );
         t1 = (u8)(op1 & 0x000f);
         sprintf( COPformat, "ELD RPD1.%d, %s", adr2, rpui[t1] );
         break;
      case 67: /* ESD0 ns,#imm:4 */
         t1 = (u8)((op1 & 0x0030) >> 4 );
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ESD0 %s, #%d", ns[t1], imm4 );
         break;
      case 68: /* ESD1 ns,#imm:4 */
         t1 = (u8)((op1 & 0x0030) >> 4 );
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ESD1 %s, #%d", ns[t1], imm4 );
         break;
      case 69: /* ESD2 ns,#imm:4 */
         t1 = (u8)((op1 & 0x0030) >> 4 );
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ESD2 %s, #%d", ns[t1], imm4 );
         break;
      case 70: /* ESD3 ns,#imm:4 */
         t1 = (u8)((op1 & 0x0030) >> 4 );
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ESD3 %s, #%d", ns[t1], imm4 );
         break;
      case 71: /* ELD An,rpdi.adr:5 */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         t2 = (u8)((op1 & 0x0080) >> 7 );
         adr5 = (u8)(op1 & 0x001f);
         sprintf( COPformat, "ELD %s, %s.%d", An[t1], rpdi[t2], adr5 );
         break;
      case 72: /* ELD rpdi.adr:5,An */
         t1 = (u8)((op1 & 0x0080) >> 7 );
         adr5 = (u8)(op1 & 0x001f);
         t2 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ELD %s.%d, %s", rpdi[t1], adr5, An[t2] );
         break;
      case 73: /* EMOD0 An,mg */
         t1 = (u8)((op1 & 0x0180) >> 7 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x001f);
         sprintf( COPformat, "%s %s, %s", emod0[t1], An[t2], mg[t3] );
         break;
      case 74: /* EMOD0 An,Am */
         t1 = (u8)((op1 & 0x000c) >> 2 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "%s %s, %s", emod0[t1], An[t2], An[t3] );
         break;
      case 75: /* EMOD0 An,mgx */
         t1 = (u8)((op1 & 0x000c) >> 2 );
         t2 = (u8)((op1 & 0x0060) >> 5 );
         t3 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "%s %s, %s", emod0[t1], An[t2], mgx[t3] );
         break;
      case 76: /* ELD mg,An */
         t1 = (u8)(op1 & 0x001f);
         t2 = (u8)((op1 & 0x0060) >> 5 );
         sprintf( COPformat, "ELD %s, %s", mg[t1], An[t2] );
         break;
      case 77: /* EMAD Mi, XiYi, Ai,Mj */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMAD %s, %s, %s, %s", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 78: /* EMSB Mi, XiYi, Ai,Mj */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMSB %s, %s, %s, %s", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 79: /* EMLD Mi, XiYi, Ai,Mj */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMLD %s, %s, %s, %s", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 80: /* EMUL XiYi, Ai,Mj */
         t1 = (u8)(op1 & 0x0003);
         t2 = (u8)((op1 & 0x0008) >> 3 );
         t3 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMUL %s, %s, %s", XiYi[t1], Ai[t2], Mi[t3] );
         break;
      case 81: /* EMAX Ai,Ci */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMAX %s, %s", Ai[t1], Ci[t2] );
         break;
      case 82: /* EMIN Ai,Ci */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMIN %s, %s", Ai[t1], Ci[t2] );
         break;
      case 83: /* EMAX Ai,Ai */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)(t1 ^ 1);
         sprintf( COPformat, "EMAX %s, %s", Ai[t1], Ai[t2] );
         break;
      case 84: /* EMIN Ai,Ai */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)(t1 ^ 1);
         sprintf( COPformat, "EMIN %s, %s", Ai[t1], Ai[t2] );
         break;
      case 85: /* NOP */
         sprintf( COPformat, "NOP" );
         break;
      case 86: /* ELD mg1d,mg1s */
         t1 = (u8)(op1 & 0x0007);
         t2 = (u8)((op1 & 0x0038) >> 3 );
         sprintf( COPformat, "ELD %s, %s", mg1[t1], mg1[t2] );
         break;
      case 87: /* ELD mg2d,mg2s */
         t1 = (u8)(op1 & 0x0007);
         t2 = (u8)((op1 & 0x0038) >> 3 );
         sprintf( COPformat, "ELD %s, %s", mg2[t1], mg2[t2] );
         break;
      case 88: /* EMAD Mi, XiYi, Ai,MjSL */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMAD %s, %s, %s, %sSL", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 89: /* EMSB Mi, XiYi, Ai,MjSL */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMSB %s, %s, %s, %sSL", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 90: /* EMLD Mi, XiYi, Ai,MjSL */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMLD %s, %s, %s, %sSL", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 91: /* EMUL XiYi, Ai,MjSL */
         t1 = (u8)(op1 & 0x0003);
         t2 = (u8)((op1 & 0x0008) >> 3 );
         t3 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMUL %s, %s, %sSL", XiYi[t1], Ai[t2], Mi[t3] );
         break;
      case 92: /* EMAD Mi, XiYi */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMAD %s, %s", Mi[t1], XiYi[t2] );
         break;
      case 93: /* EMSB Mi, XiYi */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMSB %s, %s", Mi[t1], XiYi[t2] );
         break;
      case 94: /* EMAD Mi, XiYi, Ai,MjSR */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMAD %s, %s, %s, %sSR", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 95: /* EMSB Mi, XiYi, Ai,MjSR */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMSB %s, %s, %s, %sSR", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 96: /* EMLD Mi, XiYi, Ai,MjSR */
         t1 = (u8)((op1 & 0x0001) >> 4 );
         t2 = (u8)(op1 & 0x0003);
         t3 = (u8)((op1 & 0x0008) >> 3 );
         t4 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMLD %s, %s, %s, %sSR", Mi[t1], XiYi[t2], Ai[t3], Mi[t4] );
         break;
      case 97: /* EMUL XiYi, Ai,MjSR */
         t1 = (u8)(op1 & 0x0003);
         t2 = (u8)((op1 & 0x0008) >> 3 );
         t3 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMUL %s, %s, %sSR", XiYi[t1], Ai[t2], Mi[t3] );
         break;
      case 98: /* EMLD Mi, XiYi */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)((op1 & 0x0004) >> 2 );
         sprintf( COPformat, "EMLD %s, %s", Mi[t1], XiYi[t2] );
         break;
      case 99: /* EMUL XiYi */
         t1 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "EMUL %s", XiYi[t1] );
         break;
      case 100: /* ERPR rpi */
         t1 = (u8)(op1 & 0x0003);
         sprintf( COPformat, "ERPR %s", rpi[t1] );
         break;
      case 101: /* ELD An,#imm:5 */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         imm5 = sext5to32( (u8)(op1 & 0x001f) );
         sprintf( COPformat, "ELD %s, #%d", An[t1], imm5 );
         break;
      case 102: /* EADD An,#imm:5 */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         imm5 = sext5to32( (u8)(op1 & 0x001f) );
         sprintf( COPformat, "EADD %s, #%d", An[t1], imm5 );
         break;
      case 103: /* ECP An,#imm:5 */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         imm5 = sext5to32( (u8)(op1 & 0x001f) );
         sprintf( COPformat, "ECP %s, #%d", An[t1], imm5 );
         break;
      case 104: /* EMOD1 An */
         t1 = (u8)((op1 & 0x0060) >> 5 );
         t2 = (u8)((op1 & 0x0010) >> 4 );
         t3 = (u8)(op1 & 0x000f);
         sprintf( COPformat, "%s %s", emod1[t2][t3], An[t1] );
         break;
      case 105: /* ERPN rpi,An */
         t1 = (u8)(op1 & 0x0003);
         t2 = (u8)((op1 & 0x000c) >> 2 );
         sprintf( COPformat, "ERPN %s, %s", rpi[t1], An[t2] );
         break;
      case 106: /* ERPS rps */
         t1 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ERPS %s", rps[t1] );
         break;
      case 107: /* ERPD rpd */
         t1 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ERPD %s", rpd[t1] );
         break;
      case 108: /* EMOD2 Mi */
         t1 = (u8)((op1 & 0x0010) >> 4 );
         t2 = (u8)(op1 & 0x000f);
         sprintf( COPformat, "%s", emod2[t1][t2] );
         break;
      case 109: /* ETST cc T/EC3 */
         t1 = (u8)(op1 & 0x000f);
         t2 = (u8)((op1 & 0x0010) >> 4 );
         sprintf( COPformat, "ETST%s %s", cct[t1], TEC3[t2] );
         break;
      case 110: /* ER/ES bs */
         t1 = (u8)((op1 & 0x0010) >> 4 );
         t2 = (u8)(op1 & 0x000f);
         sprintf( COPformat, "%s %s", rs[t1], bs[t2] );
         break;
      case 111: /* ELD Pi,mg1 */
         t1 = (u8)((op1 & 0x0008) >> 3 );
         t2 = (u8)(op1 & 0x0007);
         sprintf( COPformat, "ELD %s, %s", Pi[t1], mg1[t2] );
         break;
      case 112: /* ELD mg1,Pi */
         t1 = (u8)(op1 & 0x0007);
         t2 = (u8)((op1 & 0x0008) >> 3 );
         sprintf( COPformat, "ELD %s, %s", mg1[t1], Pi[t2] );
         break;
      case 113: /* ELD mgx,An */
         t1 = (u8)(op1 & 0x0003);
         t2 = (u8)((op1 & 0x000c) >> 2 );
         sprintf( COPformat, "ELD %s, %s", mgx[t1], An[t2] );
         break;
      case 114: /* ELD sdid,sdis */
         t1 = (u8)(op1 & 0x0003);
         t2 = (u8)((op1 & 0x000c) >> 2 );
         sprintf( COPformat, "ELD %s, %s", sdi[t1], sdi[t2] );
         break;
      case 115: /* EBK #imm:4 */
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "EBK #%d", imm4 );
         break;
      case 116: /* ESEC0 #imm:4 */
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ESEC0 #%d", imm4 );
         break;
      case 117: /* ESEC1 #imm:4 */
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ESEC1 #%d", imm4 );
         break;
      case 118: /* ESEC2 #imm:4 */
         imm4 = sext4to32( (u8)(op1 & 0x000f) );
         sprintf( COPformat, "ESEC2 #%d", imm4 );
         break;
      default:
         sprintf( COPformat, "Error in formatCOP.c" );
         break;
   }
   return COPformat;
}

